import bz2
import gzip
import lzma
import shutil
import tarfile
from contextlib import contextmanager
from enum import Enum, auto
from pathlib import Path
from typing import IO

from socialgene.utils.logging import log


class Compression(Enum):
    bzip2 = auto()
    gzip = auto()
    xz = auto()
    uncompressed = auto()


def is_compressed(filepath: Path) -> Compression:
    """
    Determines the compression type of a file based on its signature.

    Args:
      filepath (Path): The `filepath` parameter is a `Path` object that represents the path to the file
    that you want to check for compression.

    Returns:
      The function `is_compressed` returns the type of compression used for the file specified by the
    `filepath` parameter. The possible return values are `Compression.gzip`, `Compression.bzip2`,
    `Compression.xz`, or `Compression.uncompressed`.
    """
    with open(filepath, "rb") as f:
        signature = f.peek(8)[:8]
        if tuple(signature[:2]) == (0x1F, 0x8B):
            return Compression.gzip
        elif tuple(signature[:3]) == (0x42, 0x5A, 0x68):
            return Compression.bzip2
        elif tuple(signature[:7]) == (0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00, 0x00):
            return Compression.xz
        else:
            return Compression.uncompressed


def gunzip(filepath: Path) -> None:
    """
    The function `gunzip` decompresses a gzip file and saves the decompressed file with the same name
    but without the ".gz" extension.

    Args:
      filepath (Path): Path to the compressed file that you want to
    decompress.
    """
    if is_compressed(filepath).name == "gzip":
        # remove ".gz" for the new filepath
        new_path = Path(filepath.parents[0], filepath.stem)
        log.info(f"Started decompressing: {str(Path(filepath))}")
        # open gz, decompress, write back out to new file
        with gzip.open(filepath, "rb") as f_in:
            with open(new_path, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        filepath.unlink()
        log.info(f"Finished decompressing: {str(Path(filepath))}")
        return new_path


@contextmanager
def open_read(filepath: Path) -> IO:
    """
    The function `open_read` opens a file for reading, taking into account different compression
    formats.

    Args:
      filepath (Path): The `filepath` parameter is the path to the file that you want to open and read.
    It should be a string representing the file path.
    """
    filepath_compression = is_compressed(filepath)
    if filepath_compression == Compression.gzip:
        f = gzip.open(filepath, "rt")
    elif filepath_compression == Compression.bzip2:
        f = bz2.open(filepath, "rt")
    elif filepath_compression == Compression.xz:
        f = lzma.open(filepath, "rt")
    else:
        f = open(filepath, "rt")
    try:
        yield f
    finally:
        f.close()


@contextmanager
def open_write(filepath: str, mode="w", compression: str = None) -> IO:
    """Open a file for writing

    Args:
        filepath (str): input filepath
        mode (str, optional): modes to open file for writing (only use "a", "w", or "r"- "b" will auto-applied). Defaults to "w".
        compression (str, optional): which compression method to use ("gzip", "bzip", "xz", or None). Defaults to None.
    Yields:
        Iterator[IO]: context manager
    """
    filepath = Path(filepath)
    match compression:
        case "gzip":
            _open = gzip.open(filepath.with_suffix(filepath.suffix + ".gz"), f"{mode}")
        case "bzip":
            _open = bz2.open(filepath.with_suffix(filepath.suffix + ".bz2"), f"{mode}")
        case "xz":
            _open = lzma.open(filepath.with_suffix(filepath.suffix + ".xz"), f"{mode}")
        case None:
            _open = open(filepath, mode)
        case _:
            raise ValueError(
                f'compression variable must be "gzip", "bzip", "xz", or None; was {compression}'
            )
    try:
        yield _open
    finally:
        _open.close()


def check_if_tar(filepath):
    return tarfile.is_tarfile(filepath)


def guess_filetype(filepath):
    """Guess what type of file it is
    Args:
        filepath: file path of file to guess
    Returns:
      a string indicating the type of file. The possible return values are "genbank", "fasta", "gff", or
    "domtblout".
    """
    with open_read(filepath) as f:
        l1 = f.readline()
    if l1.startswith("LOCUS "):
        return "genbank"
    if l1.startswith(">"):
        return "fasta"
    if l1.startswith("##gff-version"):
        return "gff"
    if (
        l1.replace(" ", "")
        == "#---fullsequence-----------------thisdomain-------------hmmcoordalicoordenvcoord\n"
    ):
        return "domtblout"
    if l1.startswith("##gff-version 3"):
        return "gff"



from socialgene.base.molbio import LocusAssemblyMetadata
from socialgene.neo4j.neo4j_element import Node

# This file can only contain class objects that inherit from Node


class PARAMETERS(Node):
    """Parameters and environmental variables used during database creation"""

    neo4j_label = ["parameters"]
    description = "Parameters and environmental variables used during database creation"
    header_filename = "parameters.header"
    target_subdirectory = "parameters"
    target_extension = "socialgene_parameters"
    header = [
        "uid:ID(when)",
        "SG_LOC_NEO4J",
        "SG_LOC_HMMS",
        "NEO4J_dbms_memory_pagecache_size",
        "NEO4J_dbms_memory_heap_initial__size",
        "NEO4J_dbms_memory_heap_max__size",
        "HMMSEARCH_IEVALUE",
        "HMMSEARCH_BACKGROUND",
        "HMMSEARCH_BIASFILTER",
        "HMMSEARCH_NULL2",
        "HMMSEARCH_SEED",
        "HMMSEARCH_Z",
        "HMMSEARCH_DOMZ",
        "HMMSEARCH_F1",
        "HMMSEARCH_F2",
        "HMMSEARCH_F3",
        "HMMSEARCH_E",
        "HMMSEARCH_DOME",
        "HMMSEARCH_INCE",
        "HMMSEARCH_INCDOME",
        "HMMSEARCH_BITCUTOFFS",
        "platform",
        "architecture",
        "py_executable",
        "py_version",
        "genome_download_command",
    ]
    property_specification = {
        "uid": str,
        "SG_LOC_NEO4J": str,
        "SG_LOC_HMMS": str,
        "NEO4J_dbms_memory_pagecache_size": str,
        "NEO4J_dbms_memory_heap_initial__size": str,
        "NEO4J_dbms_memory_heap_max__size": str,
        "HMMSEARCH_IEVALUE": str,
        "HMMSEARCH_BACKGROUND": str,
        "HMMSEARCH_BIASFILTER": str,
        "HMMSEARCH_NULL2": str,
        "HMMSEARCH_SEED": int,
        "HMMSEARCH_Z": int,
        "HMMSEARCH_DOMZ": int,
        "HMMSEARCH_F1": float,
        "HMMSEARCH_F2": float,
        "HMMSEARCH_F3": float,
        "HMMSEARCH_E": float,
        "HMMSEARCH_DOME": float,
        "HMMSEARCH_INCE": float,
        "HMMSEARCH_INCDOME": float,
        "HMMSEARCH_BITCUTOFFS": str,
        "platform": str,
        "architecture": str,
        "py_executable": str,
        "py_version": str,
        "genome_download_command": str,
    }
    constraints_unique = ["uid"]


class ASSEMBLY(Node):
    """Represents a single genome/assembly/BGC. If the input was a FASTA file or if assembly wasn't in the genbank metadata then this will represent the file the data came from."""

    neo4j_label = ["assembly"]
    description = "Represents a single genome/assembly/BGC. If the input was a FASTA file or if assembly wasn't in the genbank metadata then this will represent the file the data came from."
    header_filename = "assembly.header"
    target_subdirectory = "genomic_info"
    target_extension = "assemblies"
    header = ["uid:ID(assembly)"] + sorted(LocusAssemblyMetadata.__slots__)
    constraints_unique = ["uid"]
    property_specification = {"uid": str} | {
        k: str for k in LocusAssemblyMetadata.__slots__
    }
    required_properties = ["uid"]


class NUCLEOTIDE(Node):
    """Represents a single nucleotide sequence (e.g. a contig/scaffold/chromosome)"""

    neo4j_label = ["nucleotide"]
    description = (
        "Represents a single nucleotide sequence (e.g. a contig/scaffold/chromosome)"
    )
    header_filename = "locus.header"
    target_subdirectory = "genomic_info"
    target_extension = "loci"
    header = ["uid:ID(nucleotide)"] + ["external_id"] + LocusAssemblyMetadata.__slots__
    constraints_unique = ["uid"]
    nonunique_index = ["external_id"]
    property_specification = {"uid": str, "external_id": str} | {
        k: str for k in LocusAssemblyMetadata.__slots__
    }
    required_properties = ["uid", "external_id"]


class PROTEIN(Node):
    """Represents a non-redundant protein"""

    neo4j_label = ["protein"]
    description = "Represents a non-redundant protein"
    header_filename = "protein_ids.header"
    target_subdirectory = "protein_info"
    target_extension = "protein_ids"
    constraints_unique = ["uid"]
    required_properties = ["uid"]

    def __init__(
        self,
        include_sequences=True,
    ):
        super().__init__()
        if include_sequences:
            self.header = ["uid:ID(protein)", "crc64", "sequence"]
        else:
            self.header = ["uid:ID(protein)", "crc64"]

    property_specification = {"uid": str, "crc64": str, "sequence": str}


class GOTERM(Node):
    """Represent a GO term"""

    neo4j_label = ["goterm"]
    description = "Represent a GO term"
    header_filename = "goterms.header"
    target_subdirectory = "goterms"
    target_extension = "goterms"
    header = ["uid:ID(goterm)", "name", "namespace", "def"]
    property_specification = {"uid": str, "name": str, "namespace": str}
    constraints_unique = ["uid"]


class TIGRFAM_ROLE(Node):
    """Represents a TIGRFAM role"""

    neo4j_label = ["tigrfam_role"]
    description = "Represents a TIGRFAM role"
    header_filename = "tigrfam_role.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfam_role"
    header = ["uid:ID(tigrfam_role)"]
    property_specification = {"uid": str}
    constraints_unique = ["uid"]


class TIGRFAM_MAINROLE(Node):
    """Represents a TIGRFAM main role"""

    neo4j_label = ["tigrfam_mainrole"]
    description = "Represents a TIGRFAM main role"
    header_filename = "tigrfam_mainrole.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfam_mainrole"
    header = ["uid:ID(tigrfam_mainrole)"]
    property_specification = {"uid": str}
    constraints_unique = ["uid"]


class TIGRFAM_SUBROLE(Node):
    """Represents a TIGRFAM sub role"""

    neo4j_label = ["tigrfam_subrole"]
    description = "Represents a TIGRFAM sub role"
    header_filename = "tigrfam_subrole.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfam_subrole"
    header = ["uid:ID(tigrfam_subrole)"]
    property_specification = {"uid": str}
    constraints_unique = ["uid"]


class TAXID(Node):
    """Represents a single taxon within NCBI taxonomy"""

    neo4j_label = ["taxid"]
    description = "Represents a single taxon within NCBI taxonomy"
    header_filename = "taxid.header"
    target_subdirectory = "taxdump_process"
    target_extension = "nodes_taxid"
    header = ["uid:ID(taxid)", "name", "rank"]
    constraints_unique = ["uid"]
    property_specification = {"uid": str, "name": str, "rank": str}
    required_properties = ["uid"]


class HMM_SOURCE(Node):
    """Represents the source of an HMM model (e.g. PFAM)"""

    neo4j_label = ["hmm_source"]
    description = "Represents the source of an HMM model (e.g. PFAM)"
    header_filename = "hmm_source.header"
    target_subdirectory = "hmm_info"
    target_extension = "hmminfo"
    header = [
        "uid:ID(hmm_source)",
        ":LABEL",
        "rel_path:String",
        "name:String",
        "acc:String",
        "notes:String",
        "description:String",
        "date:String",
        "hash:String",
        "hash_used:String",
        "model_length:String",
        "super_category:String",
        "category:String",
        "subcategory:String",
        "ga:String",
        "tc:String",
        "nc:String",
    ]
    property_specification = {
        "uid": str,
        ":LABEL": str,
        "rel_path": str,
        "name": str,
        "acc": str,
        "notes": str,
        "description": str,
        "date": str,
        "hash": str,
        "hash_used": str,
        "model_length": str,
        "super_category": str,
        "category": str,
        "subcategory": str,
        "ga": str,
        "tc": str,
        "nc": str,
    }
    constraints_unique = ["uid"]


class HMM(Node):
    """Represents a single non-redundant HMM model"""

    neo4j_label = ["hmm"]
    description = "Represents a single non-redundant HMM model"
    header_filename = "sg_hmm_nodes.header"
    target_subdirectory = "hmm_info"
    target_extension = "sg_hmm_nodes"
    header = ["uid:ID(hmm)"]
    constraints_unique = ["uid"]
    property_specification = {"uid": str}
    constraints_unique = ["uid"]



from socialgene.neo4j.neo4j_element import Node


class NPClassifierClass(Node):
    neo4j_label = ["npclassifier_class"]
    description = "Represents a NPClassifier class"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]


class NPClassifierPathway(Node):
    neo4j_label = ["npclassifier_pathway"]
    description = "Represents a NPClassifier pathway"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]


class NPClassifierSuperclass(Node):
    neo4j_label = ["npclassifier_superclass"]
    description = "Represents a NPClassifier superclass"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]



from collections import OrderedDict
from uuid import uuid4

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.SeqRecord import SeqRecord

import socialgene.hashing.hashing as hasher
from socialgene.base.locus_metadata import LocusAssemblyMetadata
from socialgene.config import env_vars
from socialgene.neo4j.neo4j import GraphDriver
from socialgene.nextflow.nodes import ASSEMBLY as ASSEMBLY_NODE
from socialgene.nextflow.nodes import HMM as HMM_NODE
from socialgene.nextflow.nodes import NUCLEOTIDE as NUCLEOTIDE_NODE
from socialgene.nextflow.nodes import PROTEIN as PROTEIN_NODE
from socialgene.nextflow.nodes import TAXID as TAXID_NODE
from socialgene.nextflow.relationships import ANNOTATES as ANNOTATES_REL
from socialgene.nextflow.relationships import ASSEMBLES_TO as ASSEMBLES_TO_REL
from socialgene.nextflow.relationships import ENCODES as ENCODES_REL
from socialgene.nextflow.relationships import IS_TAXON as IS_TAXON_REL
from socialgene.utils.logging import log
from socialgene.utils.simple_math import find_exp


class ProteinSequence:
    """Class used for working with protein sequences and can be initialized with either a sequence or a uid."""

    _amino_acids = [
        "A",
        "R",
        "N",
        "D",
        "C",
        "Q",
        "E",
        "G",
        "H",
        "I",
        "L",
        "K",
        "M",
        "F",
        "P",
        "S",
        "T",
        "W",
        "Y",
        "V",
        "X",
        "Z",
        "J",
        "U",
        "B",
        "O",
        "*",
    ]
    __slots__ = ["uid", "__crc64", "__md5", "_seq_len", "sequence"]

    def __init__(self, sequence: str = None, uid: str = None, seq_len: int = None):
        """Class for holding an amino acid sequence (protein)

        Args:
            sequence (str, optional): amino acid sequence
            uid (str, optional): a uid can be provided if a sequence isn't

        Raises:
            ValueError: Must provide either a sequence or a uid
        """
        # the sequence input variable has a default so that domtblout can create a socialgene object
        self.sequence = sequence
        if isinstance(self.sequence, type(None)):
            if isinstance(uid, type(None)):
                raise ValueError("Must provide either a sequence or a uid")
            else:
                self.uid = uid
        else:
            self._assign_hash()
        if seq_len:
            self._seq_len = seq_len

    @property
    def seq_len(self):
        """Return the length of the protein

        Returns:
            int: number of amino acids
        """
        return len(self.sequence)

    @seq_len.setter
    def seq_len(self, value):
        if isinstance(value, int):
            self._seq_len = value
        else:
            self._seq_len = len(self.sequence)

    def all_attributes(self):
        """
        The function returns a dictionary containing the attributes of an object.

        Returns:
          The `__dict__` method is returning a dictionary that contains the names and values of all the
        attributes of the object. The attributes are obtained using the `getattr` function and are filtered
        to only include attributes that are defined in the `__slots__` list and that exist in the object.
        The dictionary is sorted based on the names of the attributes.
        """
        return {s: getattr(self, s) for s in sorted(self.__slots__) if hasattr(self, s)}

    def _one_letter_amino_acids(self):
        """Create an ordered dictionary of amino acids. Used to count AAs in a protein sequence.

        Returns:
            dict: amino acid count
        """
        return OrderedDict({i: 0 for i in self._amino_acids})

    def _amino_acid_count(self):
        """Create a '-' separated string of an amino acid count

        Returns:
            str: sequence of amino acids
        """
        if self.sequence is None:
            return "-".join(["0" for i in self._one_letter_amino_acids()])
        else:
            return "-".join(
                [str(self.sequence.count(i)) for i in self._one_letter_amino_acids()]
            )

    def _assign_hash(self):
        """
        The function assigns a hash value to a sequence of amino acids.
        """
        self._standardize_sequence()
        self.uid = hasher.hash_aminos(self.sequence)

    @property
    def crc64(self):
        return hasher.hash_aminos(self.sequence, algo="crc64")

    @property
    def md5(self):
        return hasher.hash_aminos(self.sequence, algo="md5")

    def _standardize_sequence(self):
        """
        The function converts the protein sequence to uppercase and checks if all characters are valid
        amino acids, raising an error if an unknown character is found.
        """
        self.sequence = self.sequence.upper()
        self.sequence = self.sequence.replace(" ", "")
        self.sequence = self.sequence.strip("*")
        if not all([i in self._amino_acids for i in set(self.sequence)]):
            log.error(self.sequence)
            raise ValueError("Unknown character/letter in protein sequence")


class Location:
    __slots__ = ["start", "end", "strand"]
    # TODO: handle zero indexing

    def __init__(
        self,
        start: int = None,
        end: int = None,
        strand: int = None,
        **kwargs,  # this kwarg isn't accessed but is here so that calling Location with dict unpacking with extra args doesn't fail
    ):
        """Class describing genomic coordinates and strand direction.

        Args:
            start (int, optional): start coordinate
            end (int, optional): end coordinate
            strand (int, optional): DNA strand
        """
        self.start = start
        self.end = end
        self.strand = strand

    def all_attributes(self):
        return {s: getattr(self, s) for s in sorted(self.__slots__) if hasattr(self, s)}


class Domain:
    """Class for holding information about a domain/motif annotation"""

    __slots__ = [
        "hmm_id",
        "env_from",  # page 38 of the HMMER User Guide (http://eddylab.org/software/hmmer/Userguide.pdf) suggests using envelope (not hmm_from/to, ali_from/to)
        "env_to",
        "seq_pro_score",
        "evalue",
        "i_evalue",
        "domain_bias",
        "domain_score",
        "seq_pro_bias",
        "hmm_from",
        "hmm_to",
        "ali_from",
        "ali_to",
        "exponentialized",
    ]

    def __init__(
        self,
        hmm_id: str = None,
        env_from: int = None,
        env_to: int = None,
        seq_pro_score: float = None,
        evalue: float = None,
        i_evalue: float = None,
        domain_bias: float = None,
        domain_score: float = None,
        seq_pro_bias: float = None,
        hmm_from: int = None,
        hmm_to: int = None,
        ali_from: int = None,
        ali_to: int = None,
        exponentialized: bool = True,
        **kwargs,  # this kwarg isn't accessed but is here so that calling Domain with dict unpacking with extra args doesn't fail
    ):
        """
        Class for holding information about a domain/motif annotation

        Args:
          hmm_id (str): The `hmm_id` parameter is a string that represents the identifier of the hidden
        Markov model (HMM) associated with the domain.
          env_from (int): The `env_from` parameter represents the starting position of the domain in the
        target sequence.
          env_to (int): The `env_to` parameter represents the end position of the environment (sequence)
        in the domain. It is an integer value.
          seq_pro_score (float): The `seq_pro_score` parameter is a floating-point number that
        represents the sequence profile score of the domain.
          evalue (float): The `evalue` parameter is a floating-point number that represents the E-value
        of the domain. The E-value is a statistical measure that indicates the expected number of
        domains with a similar score or better that would occur by chance in a database of the same
        size.
          i_evalue (float): The `i_evalue` parameter is a floating-point number that represents the
        independent E-value of a domain. It is used to assess the statistical significance of the match
        between the domain and the sequence.
          domain_bias (float): The `domain_bias` parameter is a float that represents the bias score of
        a domain. It is used to measure the likelihood that a domain is present in a sequence due to
        chance rather than functional significance.
          domain_score (float): The `domain_score` parameter is a float that represents the score of the
        domain.
          seq_pro_bias (float): The `seq_pro_bias` parameter is a float value that represents the
        sequence profile bias of a domain. It is used to measure the bias in the sequence profile
        alignment.
          hmm_from (int): The `hmm_from` parameter represents the starting position of the domain in the
        HMM (Hidden Markov Model) profile. It is an integer value.
          hmm_to (int): The `hmm_to` parameter is an integer that represents the ending position of the
        hidden Markov model (HMM) alignment in the domain.
          ali_from (int): The `ali_from` parameter represents the starting position of the alignment in
        the sequence alignment. It is an integer value.
          ali_to (int): The `ali_to` parameter represents the ending position of the alignment in the
        sequence.
          exponentialized (bool): A boolean flag indicating whether the evalue and i_evalue should be
        exponentialized or not. If set to True, the evalue and i_evalue will be converted to exponential
        form. Defaults to True
        """
        super(Domain, self).__init__()
        if not isinstance(exponentialized, bool):
            raise ValueError(
                f"exponentialized mus be bool, was {type(exponentialized)}"
            )
        self.exponentialized = exponentialized
        if exponentialized:
            self.evalue = find_exp(evalue)
            self.i_evalue = find_exp(i_evalue)
        else:
            self.evalue = round(float(evalue), 1)
            self.i_evalue = round(float(i_evalue), 1)
        # some of are rounded because of differences between pyhmmer and hmmer results
        self.hmm_id = str(hmm_id)
        self.env_from = int(env_from)
        self.env_to = int(env_to)
        self.seq_pro_score = round(float(seq_pro_score), 1)
        self.domain_bias = round(float(domain_bias), 1)
        self.domain_score = round(float(domain_score), 1)
        self.seq_pro_bias = round(float(seq_pro_bias), 1)
        self.hmm_from = int(hmm_from)
        self.hmm_to = int(hmm_to)
        self.ali_from = int(ali_from)
        self.ali_to = int(ali_to)

    def all_attributes(self):
        ord_dict = OrderedDict()
        ord_dict["hmm_id"] = str(self.hmm_id)
        ord_dict["env_from"] = int(self.env_from)
        ord_dict["env_to"] = int(self.env_to)
        ord_dict["seq_pro_score"] = round(self.seq_pro_score, 1)
        ord_dict["evalue"] = round(self.evalue, 1)
        ord_dict["i_evalue"] = round(self.i_evalue, 1)
        ord_dict["domain_bias"] = round(self.domain_bias, 1)
        ord_dict["domain_score"] = round(self.domain_score, 1)
        ord_dict["seq_pro_bias"] = round(self.seq_pro_bias, 1)
        ord_dict["hmm_from"] = int(self.hmm_from)
        ord_dict["hmm_to"] = int(self.hmm_to)
        ord_dict["ali_from"] = int(self.ali_from)
        ord_dict["ali_to"] = int(self.ali_to)
        ord_dict["exponentialized"] = bool(self.exponentialized)
        return ord_dict

    def tsv_attributes(self):
        ord_dict = OrderedDict()
        ord_dict["hmm_id"] = str(self.hmm_id)
        ord_dict["env_from"] = int(self.env_from)
        ord_dict["env_to"] = int(self.env_to)
        ord_dict["seq_pro_score"] = round(self.seq_pro_score, 1)
        ord_dict["evalue"] = round(self.evalue, 1)
        ord_dict["i_evalue"] = round(self.i_evalue, 1)
        ord_dict["domain_bias"] = round(self.domain_bias, 1)
        ord_dict["domain_score"] = round(self.domain_score, 1)
        ord_dict["seq_pro_bias"] = round(self.seq_pro_bias, 1)
        ord_dict["hmm_from"] = int(self.hmm_from)
        ord_dict["hmm_to"] = int(self.hmm_to)
        ord_dict["ali_from"] = int(self.ali_from)
        ord_dict["ali_to"] = int(self.ali_to)
        ord_dict["exponentialized"] = bool(self.exponentialized)
        return ord_dict

    def get_hmm_id(self):
        return self.hmm_id

    @property
    def domain_within_threshold(self):
        """Check if a protein's domain annotation is within the threshold for inclusion (currently i_evalue based)
        Returns:
            bool: is less than or equal to the set ievalue cutoff?
        """
        if self.exponentialized:
            return self.i_evalue <= find_exp(env_vars["HMMSEARCH_IEVALUE"])
        else:
            return self.i_evalue <= env_vars["HMMSEARCH_IEVALUE"]

    def __hash__(self):
        """Used to prevent adding duplicate domains
        Returns:
            hash: hash for set()
        """
        return hash((self.hmm_id, self.env_from, self.env_to, self.i_evalue))

    def __eq__(self, other):  # pragma: no cover
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.hmm_id == other.hmm_id
            and self.env_from == other.env_from
            and self.env_to == other.env_to
            and self.i_evalue == other.i_evalue
        )


class Protein(
    ProteinSequence,
):
    """Container class for describing a single protein"""

    # seqlen is included as a variable so it can be provided (e.g. a sequence isn't provided and proteins are created manually)
    __slots__ = ["description", "external_id", "domains"]

    def __init__(
        self,
        description: str = None,
        external_id: str = None,
        domains: set = None,
        *args,
        **kwargs,
    ):
        """_summary_

        Args:
            description (str, optional): Protein description.
            external_id (str, optional): Non-hash-id descriptor (usually a database accession, e.g. NCBI's).
            seqlen (int, optional): Amino acid sequence length
            domains (Set, optional): Set of Domain() objects.
        """
        super().__init__(*args, **kwargs)
        self.description = description
        self.external_id = external_id
        self.domains = domains if domains is not None else set()

    def __eq__(self, other):  # pragma: no cover
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.uid == other.uid

    def __hash__(self):
        return hash(self.uid)

    def all_attributes(self):
        return {
            s: getattr(self, s) for s in sorted(self.__slots__) if hasattr(self, s)
        } | {"seq_len": self.seq_len}

    def add_domain(
        self,
        *args,
        **kwargs,
    ):
        """Add a domain to a protein if it meets threshold (ievalue)

        Returns:
            bool: used for counting # of domains added to protein
        """
        self.domains.add(Domain(*args, **kwargs))

    def add_domains_from_neo4j(self):
        try:
            with GraphDriver() as db:
                for i in db.run(
                    """
                    MATCH (p1:protein)<-[a1:ANNOTATES]-(h1:hmm)
                    where p1.uid in ["_xvHFg1H1f43WxPcZT1P8Qbcx60chSxh"]
                    RETURN apoc.map.merge({hmm_id:h1.uid}, properties(a1))
                    """,
                    uid=str(self.uid),
                ):
                    self.add_domain(**i.value())
        except Exception:
            log.debug(f"Error trying to retrieve domains for {self.uid}")

    def add_sequences_from_neo4j(self):
        try:
            with GraphDriver() as db:
                for i in db.run(
                    """
                    MATCH (p1:protein {uid: $uid})
                    RETURN p1.sequence as sequence
                    """,
                    uid=str(self.uid),
                ):
                    if x := i.value():
                        self.sequence = x
        except Exception:
            log.debug(f"Error trying to retrieve sequences for {self.uid}")

    @property
    def domain_list_sorted_by_mean_envelope_position(self):
        # (ie can't sort a set())
        # 'if' is to check whether domains is None
        # A tuple is passed so that the HMM annnotations are first sorted by the envelope midpoint
        # then ties are sorted alphabetically based on hmm_id for consistency
        return sorted(
            list(self.domains),
            key=lambda tx: ((tx.env_from + tx.env_to) / 2, tx.hmm_id),
        )

    @property
    def domain_vector(
        self,
    ):
        """Get the domain uids for a protein as an ordered list

        list: list of domain uids
        """
        if not self.domains:
            log.debug(f"Tried to get domains from domain-less protein {self}")
            return []
        return [
            i.get_hmm_id() for i in self.domain_list_sorted_by_mean_envelope_position
        ]

    def filter_domains(self):
        """Prune all domains in all proteins that don't meet the inclusion threshold (currently HMMER's i_evalue)"""
        _before_count = len(self.domains)
        temp = []
        self.domains = {i for i in self.domains if i.domain_within_threshold}

        del temp
        log.debug(
            f"Removed {str(_before_count - len(self.domains))} domains from {self.external_id}"
        )

    @property
    def fasta_string_defline_uid(self):
        return f">{self.uid}\n{self.sequence}\n"

    @property
    def fasta_string_defline_external_id(self):
        return f">{self.external_id}\n{self.sequence}\n"

    def _neo4j_node(self, include_sequences=True) -> PROTEIN_NODE:
        n = PROTEIN_NODE()
        dict_to_add = {"uid": self.uid, "crc64": self.crc64}
        if include_sequences:
            dict_to_add["sequence"] = self.sequence
        n.fill_from_dict(dict_to_add)
        return n

    def add_to_neo4j(self, include_sequences=True):
        n = self._neo4j_node(include_sequences=include_sequences)
        n.add_to_neo4j(create=False)
        for i in self.domains:
            hmmnode = HMM_NODE()
            hmmnode.fill_from_dict({"uid": i.hmm_id})
            rel = ANNOTATES_REL(start=hmmnode, end=self._neo4j_node())
            rel.fill_from_dict(
                {
                    k: v
                    for k, v in i.all_attributes().items()
                    if k in ANNOTATES_REL.property_specification
                }
            )
            rel.add_to_neo4j(create=False)


class Feature(Location):
    """Container class for describing a feature on a locus"""

    __slots__ = [
        "parent",
        "uid",
        "external_id",
        "type",
        "locus_tag",
        "description",
        "note",
        "goterms",
        "partial_on_complete_genome",
        "missing_start",
        "missing_stop",
        "internal_stop",
        "partial_in_the_middle_of_a_contig",
        "missing_N_terminus",
        "missing_C_terminus",
        "frameshifted",
        "too_short_partial_abutting_assembly_gap",
        "incomplete",
        "protein",
    ]

    def __init__(
        self,
        parent=None,
        uid: str = None,
        external_id: str = None,
        type: str = None,
        locus_tag: str = None,
        description=None,
        note=None,
        goterms=None,
        partial_on_complete_genome=None,
        missing_start=None,
        missing_stop=None,
        internal_stop=None,
        partial_in_the_middle_of_a_contig=None,
        missing_N_terminus=None,
        missing_C_terminus=None,
        frameshifted=None,
        too_short_partial_abutting_assembly_gap=None,
        incomplete=None,
        **kwargs,
    ):
        """
        The above function is a constructor for a class that represents a feature on a locus, with
        various attributes and optional arguments.

        Args:
          uid (str): A string representing the hash value of the protein.
          external_id (str): The unique identifier for the protein associated with the feature.
          type (str): The "type" parameter is used to specify the type of feature. In this case, it is
        used to specify the type of the feature on a locus, such as "protein".
          locus_tag (str): The locus_tag parameter is a string that represents the unique identifier for
        a feature on a locus. It is typically used in genomics to identify a specific gene or protein
        within a genome.
          description: A description of the feature on a locus.
          note: The `note` parameter is used to provide additional information or comments about the
        feature. It is an optional parameter and can be any string value.
          goterms: The `goterms` parameter is used to store the Gene Ontology terms associated with the
        feature. Gene Ontology (GO) terms are standardized terms used to describe the function,
        location, and involvement of genes and gene products in biological processes.
          partial_on_complete_genome: A boolean flag indicating whether the feature is partial on a
        complete genome.
          missing_start: A boolean indicating whether the start of the feature is missing.
          missing_stop: A boolean flag indicating whether the stop codon of the feature is missing.
          internal_stop: A boolean flag indicating whether the feature has an internal stop codon.
          partial_in_the_middle_of_a_contig: This parameter is used to indicate whether the feature is
        partial and located in the middle of a contig.
          missing_N_terminus: This parameter is used to indicate whether the N-terminus (the starting
        end) of the feature is missing. It is a boolean value, where True indicates that the N-terminus
        is missing and False indicates that it is not missing.
          missing_C_terminus: The parameter "missing_C_terminus" is used to indicate whether the feature
        is missing the C-terminus (the end) of the protein sequence. It is a boolean value that can be
        set to True or False.
          frameshifted: The "frameshifted" parameter is a boolean flag that indicates whether the
        feature has a frameshift mutation. A frameshift mutation occurs when the reading frame of a gene
        is disrupted by the insertion or deletion of nucleotides, causing a shift in the codon reading
        frame and potentially altering the amino
          too_short_partial_abutting_assembly_gap: This parameter is used to indicate whether the
        feature is too short and abuts an assembly gap.
          incomplete: A boolean flag indicating whether the feature is incomplete.
        """
        super().__init__(**kwargs)
        self.parent = parent
        self.uid = uid
        self.external_id = external_id
        self.type = type
        self.locus_tag = locus_tag
        self.description = description
        self.note = note
        self.goterms = goterms
        self.partial_on_complete_genome = partial_on_complete_genome
        self.missing_start = missing_start
        self.missing_stop = missing_stop
        self.internal_stop = internal_stop
        self.partial_in_the_middle_of_a_contig = partial_in_the_middle_of_a_contig
        self.missing_N_terminus = missing_N_terminus
        self.missing_C_terminus = missing_C_terminus
        self.frameshifted = frameshifted
        self.too_short_partial_abutting_assembly_gap = (
            too_short_partial_abutting_assembly_gap
        )
        self.incomplete = incomplete
        if self.feature_is_protein():
            # Check if there is a SocialGene object in the parent chain
            # If there is, link the feature protein attribute to the SocialGene protein object
            sg = None
            current_object = self
            for _ in range(1, 100):
                if (
                    f"{current_object.__class__.__module__}.{current_object.__class__.__name__}"
                    == "socialgene.base.socialgene.SocialGene"
                ):
                    sg = current_object
                    break
                else:
                    if hasattr(current_object, "parent"):
                        current_object = current_object.parent
            if sg:
                if self.uid in sg.proteins:
                    self.protein = sg.proteins[self.uid]

    def all_attributes(self):
        return {s: getattr(self, s) for s in sorted(self.__slots__) if hasattr(self, s)}

    def feature_is_protein(self):
        """Check if the feature "is a protein"

        Returns:
            bool: True if socialgene thinks it's a protein, False if not
        """
        # types of genbank file features to classify as a "protein"
        types_list = ["protein", "CDS"]
        return any([True for i in types_list if i == self.type])

    @property
    def fasta_string_defline_uid(self):
        try:
            if self.protein:
                return f">{self.protein.uid}\n{self.protein.sequence}\n"
        except AttributeError:
            log.debug(f"Erorr in fasta_string_defline_uid for {self}")

    @property
    def fasta_string_defline_external_id(self):
        try:
            if self.protein:
                return f">{self.protein.external_id}\n{self.protein.sequence}\n"
        except AttributeError:
            log.debug(f"Erorr in fasta_string_defline_external_id for {self}")

    def __hash__(self):
        """Used to prevent adding duplicate features to a locus (for hash in set() in Assembly.add_locus())

        Returns:
            hash: hash for set()
        """
        return hash((self.end, self.start, self.uid, self.strand, self.type))

    def __eq__(self, other):
        """Used for set() in Assembly.add_locus()"""
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.parent == other.parent
            and self.end == other.end
            and self.start == other.start
            and self.uid == other.uid
            and self.strand == other.strand
            and self.type == other.type
        )

    def __lt__(self, other):
        return self.start < other.start and self.parent == other.parent


class FeatureCollection:
    __slots__ = [
        "parent",
        "features",
    ]

    def add_feature(self, **kwargs):
        """Add a feature to a locus"""
        self.features.add(Feature(parent=self, **kwargs))

    def all_attributes(self):
        return {s: getattr(self, s) for s in sorted(self.__slots__) if hasattr(self, s)}

    @property
    def feature_uid_set(self):
        return {i.uid for i in self.features}

    @property
    def features_sorted_by_midpoint(self):
        """Sorts features by mid-coordinate of each feature"""
        for i in sorted(list(self.features), key=lambda i: int((i.end + i.start) / 2)):
            yield i

    def drop_non_protein_features(self):
        self.features = {i for i in self.features if i.type == "protein"}

    def get_feature_by_uid(self, uid):
        for k, v in self.features.items():
            if v.uid == uid:
                return v

    @property
    def proteins(self):
        sg = None
        current_object = self
        for i in range(1, 100):
            if (
                str(type(current_object))
                == "<class 'socialgene.base.socialgene.SocialGene'>"
            ):
                sg = current_object
                break
            else:
                current_object = current_object.parent
        return {
            i: sg.proteins.get(i, None)
            for i in {i.uid for i in self.features}
            if i in sg.proteins
        }

    @property
    def protein_iter(self):
        for i in self.proteins.values():
            yield i

    @property
    def fasta_string_defline_uid(self):
        for v in self.proteins.values():
            yield f">{v.uid}\n{v.sequence}\n"

    @property
    def fasta_string_defline_external_id(self):
        for v in self.proteins.values():
            yield f">{v.external_id}\n{v.sequence}\n"

    def _neo4j_node(self, **kwargs):
        n = NUCLEOTIDE_NODE()
        dict_to_add = {"uid": self.uid, "external_id": self.external_id}
        dict_to_add = dict_to_add | dict(self.metadata.all_attributes())
        n.fill_from_dict(dict_to_add)
        return n

    def add_to_neo4j(self):
        # create node
        n = self._neo4j_node()
        n.add_to_neo4j(create=False)
        protein_nodes = set()
        encodes_rels = set()
        # connect to proteins
        for i in self.features:
            protein_nodes.add(i.protein._neo4j_node())
            temp_rel = ENCODES_REL(start=n, end=i.protein._neo4j_node())
            properties = {
                k: v
                for k, v in i.all_attributes().items()
                if k in ENCODES_REL.property_specification
            }
            properties = properties | {
                "start": i.start,
                "end": i.end,
                "strand": i.strand,
            }
            temp_rel.fill_from_dict(properties)
            encodes_rels.add(temp_rel)
        if protein_nodes:
            protein_nodes = list(protein_nodes)
            protein_nodes[0].add_multiple_to_neo4j(protein_nodes, create=False)
        if encodes_rels:
            encodes_rels = list(encodes_rels)
            encodes_rels[0].add_multiple_to_neo4j(encodes_rels, create=False)
        # connect to assembly
        assembly_node = self.parent._neo4j_node()
        assembly_node.add_to_neo4j(create=False)
        ASSEMBLES_TO_REL(start=n, end=assembly_node).add_to_neo4j(create=False)

    def write_genbank(self, outpath, start=float("-inf"), end=float("inf")):
        record = SeqRecord(
            Seq(""),
            id=self.external_id,
            name=self.external_id,
            description="A GenBank file generated by SocialGene.",
            dbxrefs=[f"Assembly:{self.parent.uid}"],
        )
        # Add annotation
        for feature in self.features_sorted_by_midpoint:
            if start:
                if int(feature.start) < int(start):
                    continue
            if end:
                if int(feature.end) > int(end):
                    continue
            biofeat = SeqFeature(
                FeatureLocation(
                    start=feature.start
                    - 1,  # biopython modifies the start position by 1
                    end=feature.end,
                    strand=feature.strand,
                ),
                type=feature.type,
                qualifiers={
                    k: v
                    for k, v in feature.all_attributes().items()
                    if v and k != "parent"
                }
                | {"translation": self.protein.sequence},
            )
            record.features.append(biofeat)
        record.annotations["molecule_type"] = "DNA"
        SeqIO.write(
            record,
            outpath,
            "genbank",
        )


class Locus(FeatureCollection):
    """Container holding a set() of genomic features"""

    __slots__ = [
        "parent",
        "features",
        "metadata",
        "external_id",
        "uid",
        "gene_clusters",
    ]

    def __init__(self, parent=None, external_id=None):
        super().__init__()
        self.parent = parent
        self.features = set()
        self.metadata = LocusAssemblyMetadata()
        self.external_id = external_id
        self.uid = self.calc_uid()
        self.gene_clusters = list()

    def calc_uid(self):
        return hasher.hasher(f"{self.parent.uid}___{self.external_id}")

    def add_bgcs_by_feature(self, features, **kwargs):
        if not all([isinstance(i, Feature) for i in features]):
            raise ValueError(
                f"All features must be of type Feature, not {[type(i) for i in features if not isinstance(i, Feature)]}"
            )
        self.gene_clusters.append(GeneCluster(features, parent=self, **kwargs))

    def add_bgcs_by_start_end(self, start, end, **kwargs):
        features = {i for i in self.features if i.start >= start and i.end <= end}
        if features:
            self.add_bgcs_by_feature(features=features, **kwargs)

    def write_genbank(self, outpath, start=None, end=None):
        record = SeqRecord(
            Seq(""),
            id=self.external_id,
            name=self.external_id,
            description="A GenBank file generated by SocialGene.",
            dbxrefs=[f"Assembly:{self.parent.uid}"],
        )
        # Add annotation
        for feature in self.features_sorted_by_midpoint:
            if start:
                if int(feature.start) < int(start):
                    continue
            if end:
                if int(feature.end) > int(end):
                    continue
            if feature.type == "protein":
                feat_type = "CDS"
            else:
                feat_type = feature.type
            biofeat = SeqFeature(
                FeatureLocation(
                    start=feature.start - 1,
                    end=feature.end,
                    strand=feature.strand,
                ),
                type=feat_type,
                qualifiers={
                    k: v
                    for k, v in feature.all_attributes().items()
                    if v and k != "parent"
                }
                | {"translation": self.parent.parent.proteins[feature.uid].sequence},
            )
            record.features.append(biofeat)
        record.annotations["molecule_type"] = "DNA"
        SeqIO.write(
            record,
            outpath,
            "genbank",
        )


class GeneCluster(FeatureCollection):
    def __init__(self, features, parent=None, uid=None, tool=None, **kwargs):
        super().__init__()
        self.parent = parent
        self.features = features
        if uid:
            self.uid = uid
        else:
            self.uid = str(uuid4())
        # self.tool; e.g.antismash, gecco, etc
        self.tool = tool
        # flexible attributes
        self.__dict__.update(kwargs)

    def __hash__(self):
        return hash((self.parent, self.uid))

    def __lt__(self, other):
        return 1

    def write_genbank(self, outpath):
        record = SeqRecord(
            Seq(""),
            id=self.parent.external_id,
            name=self.parent.external_id,
            description="A GenBank file generated by SocialGene.",
            dbxrefs=[f"Assembly:{self.parent.uid}"],
        )
        # Add annotation
        for feature in self.features:
            if feature.type == "protein":
                feat_type = "CDS"
            else:
                feat_type = feature.type
            biofeat = SeqFeature(
                FeatureLocation(
                    start=feature.start - 1,
                    end=feature.end,
                    strand=feature.strand,
                ),
                type=feat_type,
                qualifiers={
                    k: v
                    for k, v in feature.all_attributes().items()
                    if v and k != "parent"
                }
                | {"translation": self.proteins[feature.uid].sequence},
            )
            record.features.append(biofeat)
        record.annotations["molecule_type"] = "DNA"
        SeqIO.write(
            record,
            outpath,
            "genbank",
        )


class Taxonomy:
    "Class is a reserved word so just underscore all ranks to be consistent"

    __slots__ = [
        "species_",
        "genus_",
        "family_",
        "order_",
        "class_",
        "phylum_",
        "clade_",
        "superkingdom_",
    ]

    def __init__(
        self,
        species_: str = None,
        genus_: str = None,
        family_: str = None,
        order_: str = None,
        class_: str = None,
        phylum_: str = None,
        clade_: str = None,
        superkingdom_: str = None,
        **kwargs,
    ) -> None:
        self.species_ = species_
        self.genus_ = genus_
        self.family_ = family_
        self.order_ = order_
        self.class_ = class_
        self.phylum_ = phylum_
        self.clade_ = clade_
        self.superkingdom_ = superkingdom_


class Assembly:
    """Container class holding a dictionary of loci (ie genes/proteins)"""

    __slots__ = [
        "parent",
        "loci",
        "taxid",
        "metadata",
        "uid",
        "taxonomy",
        "name",
    ]

    def __init__(self, uid, parent=None):
        super().__init__()
        self.parent = parent
        self.uid = uid
        self.loci = {}
        self.taxid = None
        self.taxonomy = Taxonomy()
        self.metadata = LocusAssemblyMetadata()
        self.name = uid

    @property
    def feature_uid_set(self):
        """Return all protein hashes within assembly
        Returns:
            set: protein hashes
        """
        return set().union(*[i.feature_uid_set for i in self.loci.values()])

    def all_attributes(self):
        return {s: getattr(self, s) for s in sorted(self.__slots__) if hasattr(self, s)}

    def add_locus(self, external_id: str = None):
        """Add a locus to an assembly object"""
        if external_id is None:
            external_id = str(uuid4())
        if external_id not in self.loci:
            self.loci[external_id] = Locus(parent=self, external_id=external_id)
        else:
            log.debug(f"{external_id} already present")

    def fill_taxonomy_from_db(self):
        try:
            with GraphDriver() as db:
                res = db.run(
                    """
                        MATCH (a1:assembly {uid: $uid})-[:IS_TAXON]->(t2:taxid)-[:TAXON_PARENT*1..]->(t1:taxid)
                        WHERE t1.rank in ["genus", "family", "order", "class", "phylum", "clade", "superkingdom"]
                        WITH  apoc.map.fromLists([t2.rank],[t2.name]) as a, apoc.map.fromLists([t1.rank],[t1.name]) as b
                        return  apoc.map.mergeList(collect(a)+collect(b)) as tax_dict
                        """,
                    uid=str(self.uid),
                ).value()
                # Dict comprehension appends '_' to the keys, to match the args in Taxonomy
                self.taxonomy = Taxonomy(**{f"{k}_": v for k, v in res[0].items()})
        except Exception:
            log.debug(f"Error trying to retrieve taxonomy for {self.uid}")

    def fill_properties(self):
        try:
            with GraphDriver() as db:
                res = db.run(
                    """
                        MATCH (a1:assembly {uid: $uid})
                        return properties(a1)
                        """,
                    uid=self.uid,
                ).value()[0]
                self.metadata.update(res)
        except Exception:
            log.debug(f"Error trying to retrieve taxonomy for {self.uid}")

    def get_locus_by_uid(self, uid):
        for k, v in self.loci.items():
            if v.uid == uid:
                return v

    @property
    def gene_clusters(self):
        for locus in self.loci.values():
            for gene_cluster in locus.gene_clusters:
                yield gene_cluster

    @property
    def fasta_string_defline_uid(self):
        for v in self.loci.values():
            yield v.fasta_string_defline_uid

    @property
    def fasta_string_defline_external_id(self):
        for v in self.loci.values():
            yield v.fasta_string_defline_external_id

    def _neo4j_node(
        self,
    ):
        n = ASSEMBLY_NODE()
        dict_to_add = {"uid": self.uid}
        dict_to_add = dict_to_add | {
            k: v
            for k, v in self.metadata.all_attributes().items()
            if k in ASSEMBLY_NODE.property_specification
        }
        n.fill_from_dict(dict_to_add)
        return n

    def add_to_neo4j(self, create=False):
        for i in self.loci.values():
            i.add_to_neo4j()
        if self.taxid:
            taxnode = TAXID_NODE()
            taxnode.fill_from_dict({"uid": self.taxid})
            IS_TAXON_REL(start=self._neo4j_node(), end=taxnode).add_to_neo4j(
                create=False
            )


class Molbio:
    """Class for inheriting by SocialGene()"""

    def __init__(self):
        super().__init__()  # TODO: ?

        self.assemblies = {}  # TODO: ?
        self.proteins = {}  # TODO: ?

    def get_all_feature_uids(self):
        """Return a list of all proteins uids

        Returns:
            list: List of all proteins uids
        """
        return [i.uid for i in self.proteins.values()]

    def add_protein(
        self,
        return_uid=True,
        **kwargs,
    ):
        """Add a protein to the protein dictionary

        Args:
            no_return (bool, optional): Whether the function should return the protein's uid. Defaults to False.

        Returns:
            str: Protein's hash
        """
        temp_protein = Protein(**kwargs)
        # only add protein if it doesn't already exist
        if temp_protein.uid not in self.proteins:
            # deepcopy teo ensure instances aren't shared
            self.proteins[temp_protein.uid] = temp_protein
        if return_uid:
            return temp_protein.uid

    def add_assembly(self, uid: str = None, parent=None):
        """Add an assembly to a SocialGene object

        Args:
            uid (str, optional): Assembly identifier, should be unique across parsed input. Defaults to None.
        """
        if uid is None:
            uid = str(uuid4())
        if uid not in self.assemblies:
            self.assemblies[uid] = Assembly(uid=uid, parent=parent)
        else:
            log.debug(f"{uid} already present")



from socialgene.addons.gnps_library.nr import GnpsLibrarySpectrumNode


class GNPSLibrarySpectrum:
    def __init__(self, specnet: dict):
        self.properties = {}

    def from_specnet(self, specnet: dict):
        self.properties["uid"] = str(specnet["spectrumid"])
        self.properties["compound_name"] = str(specnet["compound_name"])
        self.properties["compound_source"] = str(specnet["compound_source"])
        self.properties["pi"] = str(specnet["pi"])
        self.properties["data_collector"] = str(specnet["data_collector"])
        self.properties["adduct"] = str(specnet["adduct"])
        self.properties["precursor_mz"] = float(specnet["precursor_mz"])
        self.properties["exactmass"] = float(specnet["exactmass"])
        self.properties["charge"] = int(specnet["charge"])
        self.properties["cas_number"] = str(specnet["cas_number"])
        self.properties["pubmed_id"] = str(specnet["pubmed_id"])
        self.properties["smiles"] = str(specnet["smiles"])
        self.properties["inchi"] = str(specnet["inchi"])
        self.properties["inchi_aux"] = str(specnet["inchi_aux"])
        self.properties["library_class"] = str(specnet["library_class"])
        self.properties["ionmode"] = str(specnet["ionmode"])
        self.properties["libraryqualitystring"] = str(specnet["libraryqualitystring"])
        self.properties["mqscore"] = float(specnet["mqscore"])
        self.properties["tic_query"] = float(specnet["tic_query"])
        self.properties["rt_query"] = float(specnet["rt_query"])
        self.properties["mzerrorppm"] = float(specnet["mzerrorppm"])
        self.properties["sharedpeaks"] = int(specnet["sharedpeaks"])
        self.properties["massdiff"] = float(specnet["massdiff"])
        self.properties["libmz"] = float(specnet["libmz"])
        self.properties["specmz"] = float(specnet["specmz"])
        self.properties["speccharge"] = int(specnet["speccharge"])
        self.properties["moleculeexplorerdatasets"] = str(
            specnet["moleculeexplorerdatasets"]
        )
        self.properties["moleculeexplorerfiles"] = str(specnet["moleculeexplorerfiles"])
        self.properties["molecular_formula"] = str(specnet["molecular_formula"])
        self.properties["inchikey"] = str(specnet["inchikey"])
        self.properties["inchikey_planar"] = str(specnet["inchikey_planar"])

    @property
    def node(self):
        if self.properties:
            return GnpsLibrarySpectrumNode(properties=self.properties)



import logging
import sys

from rich.console import Console

CONSOLE = Console(width=150)
# If {rich} is installed use it, otherwise.... don't
try:
    from rich.logging import RichHandler

    # https://rich.readthedocs.io/en/stable/logging.html
    logging.basicConfig(
        level=logging.INFO,
        #  format="%(filename)s/%(module)s/%(funcName)s\::: %(message)s",
        format="%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[
            RichHandler(
                rich_tracebacks=True, tracebacks_word_wrap=False, console=CONSOLE
            )
        ],
    )

    log = logging.getLogger(__name__)

except ImportError:
    log = logging.getLogger(__name__)

# log.setLevel(env_vars["SOCIALGENE_LOGLEVEL"])
# handler = logging.StreamHandler(stream=sys.stdout)

# log.addHandler(handler)
logging.getLogger("neo4j").setLevel(logging.WARNING)
logging.getLogger("asyncio").setLevel(logging.WARNING)


def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    log.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception

# https://blog.hay-kot.dev/fastapi-and-rich-tracebacks-in-development/
# https://coralogix.com/blog/python-logging-best-practices-tips/



#
# from urllib.request import urlopen
# from pathlib import Path

#
# import xml.etree.ElementTree as ET

#


# def read_biosample_attributes_xml():

#     # read the xml format from ncbi
#     var_url = urlopen(
#         "https://www.ncbi.nlm.nih.gov/biosample/docs/attributes/?format=xml"
#     )
#     return ET.parse(var_url)


# def parse_biosample_attributes_xml(tree):
#     """Parse the returned XML

#     Args:
#         tree (xml.etree.ElementTree.ElementTree): [description]

#     Returns:
#         list:
#     """

#     root = tree.getroot()
#     attr = []
#     for i in root.iterfind("Attribute/HarmonizedName"):
#         attr.append(i.text)
#     attr = set(attr)  # make unique
#     attr = list(attr)
#     attr.sort()
#     return attr


# def write_attributes(outdir, attribute_list):
#     biosample_attribute_tags_path = Path(outdir, "biosample_attributes")
#     with open(biosample_attribute_tags_path, "w") as handle:
#         handle.write("\n".join(attribute_list))


# tree = read_biosample_attributes_xml()

# attribute_list = parse_biosample_attributes_xml(tree=tree)

# write_attributes(
#     outdir="/home/chase/Documents/socialgene/socialgene/src/socialgene/data",
#     attribute_list=attribute_list,
# )



import argparse
from pathlib import Path

from socialgene.addons.gnps_networking.parse import GNPS_SNETS

parser = argparse.ArgumentParser(
    description="Integrate a GNPS molecular network into a SocialGene Neo4j Database"
)
parser.add_argument(
    "--inputdir",
    help="Path to unzipped directory of GNPS molecular network download",
    default=False,
    required=True,
)


def main():
    args = parser.parse_args()
    if not Path(args.inputdir).exists():
        raise FileNotFoundError(f"Path {args.inputdir} does not exist")
    if not Path(args.inputdir).is_dir():
        raise ValueError(f"Path {args.inputdir} is not a directory")
    gnps = GNPS_SNETS(gnps_dirpath=args.inputdir)
    gnps.add_gnps_library_spectra()
    gnps.add_gnps_library_spectrum_classifications()
    gnps.add_gnps_library_spectrum_ion_sources()
    gnps.add_gnps_library_spectrum_instruments()
    gnps.add_gnps_library_spectrum_organisms()
    gnps.add_gnps_clusters()
    gnps.add_links_between_gnps_clusters_and_assemblies()
    gnps.add_links_between_gnps_clusters()


if __name__ == "__main__":
    main()



import argparse
from pathlib import Path

from socialgene.cli.search.sea import search_bgc

parser = argparse.ArgumentParser(
    description="Search a SocialGene database for gene clusters similar to an input gene cluster"
)

parser.add_argument(
    "--gbk_path",
    type=Path,
    help="Path to the query gene cluster GenBank-format file",
    required=True,
)
parser.add_argument(
    "--hmm_dir",
    type=str,
    help="Path to the directory containing HMM files created by SocialGene's Nextflow workflow",
    required=True,
)
parser.add_argument(
    "--clinker_outpath",
    type=Path,
    help="Path to the output Clinker JSON file",
)
parser.add_argument(
    "--assemblies_must_have_x_matches",
    type=float,
    default=0.8,
    help="Minimum query proteins an assembly must have (<1 == fraction of query proteins)",
)
parser.add_argument(
    "--nucleotide_sequences_must_have_x_matches",
    type=float,
    default=0.8,
    help="Minimum query proteins a nucleotide sequence must have (<1 == fraction of query proteins)",
)
parser.add_argument(
    "--gene_clusters_must_have_x_matches",
    type=float,
    default=0.8,
    help="Minimum query proteins a target gene cluster sequence must have (<1 == fraction of query proteins)",
)
parser.add_argument(
    "--break_bgc_on_gap_of",
    type=int,
    default=10000,
    help="Split gene clusters if gap between matched proteins is greater than this value",
)
parser.add_argument(
    "--target_bgc_padding",
    type=int,
    default=2000,
    help="Pull proteins x-nucleotides on either side of target gene cluster",
)
parser.add_argument(
    "--max_domains_per_protein",
    type=float,
    default=3,
    help="Maximum number of domains per protein used in search; 0 for all",
)
parser.add_argument(
    "--max_outdegree",
    type=int,
    help="Maximum outdegree a query domain can have; 0 for all",
)
parser.add_argument(
    "--max_query_proteins",
    type=float,
    default=10,
    help="Maximum number of query proteins to use in search; 0 for all",
)
parser.add_argument(
    "--scatter",
    help="Pull proteins from across the width of the input gene cluster",
    action=argparse.BooleanOptionalAction,
    default=False,
)
parser.add_argument(
    "--locus_tag_bypass_list",
    nargs="+",
    help="List of locus tags to bypass search filter",
)
parser.add_argument(
    "--protein_id_bypass_list",
    nargs="+",
    help="List of protein IDs to bypass search filter",
)
parser.add_argument(
    "--only_culture_collection",
    help="Only search genomes from culture collections",
    action=argparse.BooleanOptionalAction,
    default=False,
)
parser.add_argument(
    "--frac",
    type=float,
    default=0.75,
    help="Fraction of domains that equals a match in the initial database scan",
)
parser.add_argument(
    "--run_async",
    help="Run the initial database scan using asynchronously",
    action=argparse.BooleanOptionalAction,
    default=True,
)
parser.add_argument(
    "--analyze_with",
    type=str,
    default="hmmer",
    help="Tool to use for reciprocal best hit comparison; one of 'hmmer', 'blastp', 'mmseqs2'",
)


def main():
    args = parser.parse_args()

    if args.max_domains_per_protein == 0:
        args.max_domains_per_protein = None
    if args.max_outdegree == 0:
        args.max_outdegree = None
    if args.max_query_proteins == 0:
        args.max_query_proteins = None
    _ = search_bgc(
        input=args.gbk_path,
        hmm_dir=args.hmm_dir,
        outpath_clinker=args.clinker_outpath,
        use_neo4j_precalc=True,
        assemblies_must_have_x_matches=args.assemblies_must_have_x_matches,
        nucleotide_sequences_must_have_x_matches=args.nucleotide_sequences_must_have_x_matches,
        gene_clusters_must_have_x_matches=args.gene_clusters_must_have_x_matches,
        break_bgc_on_gap_of=args.break_bgc_on_gap_of,
        target_bgc_padding=args.target_bgc_padding,
        max_domains_per_protein=args.max_domains_per_protein,
        max_outdegree=args.max_outdegree,
        max_query_proteins=args.max_query_proteins,
        scatter=args.scatter,
        locus_tag_bypass_list=args.locus_tag_bypass_list,
        protein_id_bypass_list=args.protein_id_bypass_list,
        only_culture_collection=args.only_culture_collection,
        frac=args.frac,
        run_async=args.run_async,
        analyze_with=args.analyze_with,
    )


if __name__ == "__main__":
    main()



import atexit
import importlib.resources
from typing import Any

from neo4j import GraphDatabase

from socialgene.config import env_vars
from socialgene.utils.logging import CONSOLE, log


# Read in queries from cypher file, as dictionary
def import_queries():
    cypher_dictionary = {}
    with open(
        importlib.resources.files("socialgene").joinpath("neo4j", "queries.cypher"), "r"
    ) as f:
        for line in f:
            pass
            if line.startswith("// Name:"):
                query_name = line.replace("// Name:", "").strip()
                cypher_dictionary[query_name] = {
                    "query": "",
                    "description": "",
                    "param": "",
                }
            elif line.startswith("// Description:"):
                temp = line.replace("// Description:", "").strip()
                cypher_dictionary[query_name]["description"] = temp
                del temp
            elif line.startswith("// Param:"):
                temp = line.replace("// Param:", "").strip()
                cypher_dictionary[query_name]["Param"] = temp
                del temp
            else:
                cypher_dictionary[query_name][
                    "query"
                ] = f"{cypher_dictionary[query_name]['query']}\n{line.strip()}"
    return cypher_dictionary


# "Create the driver once at application start and then make it available to the application."
#  https://community.neo4j.com/t/neo4j-python-driver-service-unavailable/4111/9
class GraphDriver(object):
    """The `GraphDriver` class provides a connection to a Neo4j database and manages
    the session for executing transactions."""

    instance = None
    driver = None

    def __new__(cls):
        if cls.instance is None:
            inst = cls.instance = object.__new__(cls)
            try:
                inst.driver = GraphDatabase.driver(
                    env_vars["NEO4J_URI"],
                    auth=(
                        env_vars["NEO4J_USER"],
                        env_vars["NEO4J_PASSWORD"],
                    ),
                )
                log.info(f"Connected to Neo4j database at {env_vars['NEO4J_URI']}")
            except Exception as e:
                log.exception("Error connecting to Neo4j database")
                log.exception(e)
        atexit.register(cls.shutdown)
        return cls.instance

    def __init__(self) -> None:
        self.session = None
        # self.spinner handled by class' context management protocol
        self.spinner = CONSOLE.status(
            "Executing Neo4j transaction", spinner="bouncingBar"
        )

    def check_connection(self):
        self.driver.verify_connectivity()

    def __enter__(self):
        # self.spinner.start()
        self.session = self.driver.session()
        return self.session

    def __exit__(self, *args, **kwargs):
        # self.spinner.stop()
        self.session.close()

    @classmethod
    def shutdown(cls):
        if cls.instance:
            cls.instance.driver.close()
            cls.instance = None


class Neo4jQuery:
    """The `Neo4jQuery` class provides methods for printing and running Neo4j queries."""

    def __init__(self):
        pass

    @staticmethod
    def print_query(query_name):  # pragma: no cover
        log.info(import_queries()[query_name]["query"])

    @staticmethod
    def print_queries():  # pragma: no cover
        resolved_query_dict = import_queries()
        for k in resolved_query_dict.keys():
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            log.info(f"Query name: {k}")
            log.info(f'Description: \t{resolved_query_dict[k]["description"]}')
            log.info(f'Parameter(s): \t{resolved_query_dict[k]["param"]}')
            log.info(f'Query: {resolved_query_dict[k]["query"]}')
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

    @staticmethod
    def query_neo4j(
        cypher_name: str = None,
        cypher: str = None,
        param: Any = None,
        rettype: str = "data",
        *args,
        **kwargs,
    ):
        """Run a provided cypher query on {param}

        Args:
            cypher_name (str, optional): Neo4j Cypher query. Defaults to None.
            cypher (str, optional): Single string of Cypher script.  Defaults to None.
            param (Any, optional): parameter passed to Neo4j query, type depends on query. Defaults to None.
            rettype (str, optional): output function. For available methods see (https://neo4j.com/docs/api/python-driver/current/api.html#result). Defaults to "data"
            args/kwargs (Any): pas additional argument(s) to the rettype method
        Returns:
            Any: type depends on query
        """
        # See queries.py
        resolved_query_dict = import_queries()
        # grab the the neo4j connection
        if cypher_name:
            query = resolved_query_dict[cypher_name]["query"]
        elif cypher:
            query = cypher
        with GraphDriver() as db:
            # make the query against the db
            results = db.run(query, param=param)
            return getattr(results, rettype)(*args, **kwargs)



from socialgene.neo4j.neo4j_element import Node, Relationship
from socialgene.nextflow.nodes import NUCLEOTIDE


class GeneCluster(Node):
    neo4j_label = ["genecluster"]
    description = "A gene cluster"
    property_specifications = {
        "uid": str,
        "start": int,
        "end": int,
    }
    required_properties = ["uid", "start", "end"]


class GeneClusterToNucleotide(Relationship):
    neo4j_label = "ENCODES"
    description = "Connects the nucleotide sequence to the gene cluster"
    property_specifications = {
        "start": int,
        "end": int,
        "core_start": int,
        "core_end": int,
    }
    start_class = NUCLEOTIDE
    end_class = GeneCluster



#
# import argparse

#

#
# from socialgene.utils.logging import log
# from socialgene.parsers.ncbi_feature_table import NCBIFeatureTable
# from socialgene.config import env_vars

# parser = argparse.ArgumentParser(
#     description="Export header files for Neo4j admin import"
# )

# parser.add_argument(
#     "--input",
#     metavar="filepath",
#     help="Input feature tables",
#     required=True,
# )

# parser.add_argument(
#     "--sqlite_path",
#     metavar="filepath",
#     help="Path to SQLite database",
#     required=True,
# )
# parser.add_argument(
#     "--outdir",
#     metavar="filepath",
#     help="Output directory filepath",
#     required=True,
# )

# input_path = (
#     "/home/chase/Documents/socialgene/test_data/BGC0000001_feature_table.txt.gz"
# )


# def main():
#     args = parser.parse_args()
#     outdir = args.outdir
#     log.info(f"Socialgene variables: \n{env_vars}")
#     ft_object = NCBIFeatureTable()
#     ft_object.connect("/home/chase/Documents/socialgene/test_data/hashid.sqlite")
#     ft_object.parse(args.input)

#     ft_object.export_locus_to_protein(outdir=outdir)
#     ft_object.export_protein_info(outdir=outdir)
#     ft_object.export_assembly_to_locus(outdir=outdir)
#     ft_object.export_loci(outdir=outdir)
#     ft_object.export_assembly(outdir=outdir)


# if __name__ == "__main__":
#     main()



import csv
import pickle
import tempfile
from multiprocessing import cpu_count
from operator import attrgetter
from pathlib import Path
from typing import List

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.SeqRecord import SeqRecord
from rich.progress import Progress

from socialgene.base.compare_protein import CompareProtein
from socialgene.base.molbio import Molbio
from socialgene.hmm.hmmer import HMMER
from socialgene.neo4j.neo4j import GraphDriver, Neo4jQuery
from socialgene.neo4j.search.basic import search_protein_hash
from socialgene.parsers.hmmer_parser import HmmerParser
from socialgene.parsers.sequence_parser import SequenceParser
from socialgene.utils.chunker import chunk_a_list_with_numpy
from socialgene.utils.file_handling import open_write
from socialgene.utils.logging import log


class SocialGene(Molbio, CompareProtein, SequenceParser, Neo4jQuery, HmmerParser):
    """Main class for building, sotring, working with protein and genomic info"""

    # _export_table_names exports tables for the nextflow pipeline
    _export_table_names = [
        "table_protein_to_go",
        "table_assembly_to_locus",
        "table_loci",
        "table_locus_to_protein",
        "table_assembly",
        "table_assembly_to_taxid",
        "table_protein_ids",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.protein_comparison = []

    ########################################
    # Get info
    ########################################

    def get_loci_ids(self):
        return list(self.loci.keys())

    def get_all_gene_clusters(self):
        for i in self.assemblies.values():
            for j in i.loci.values():
                for k in j.gene_clusters:
                    yield k

    @property
    def protein_iter(self):
        for i in self.proteins.values():
            yield i

    ########################################
    # Filter
    ########################################

    def get_protein_domains_from_db(self, protein_id_list):
        # Search neo4j for a protein with the same hash, if it exists,
        # retrieve the domain info for it
        get_protein_domains_result = self.query_neo4j(
            cypher_name="get_protein_domains",
            param=protein_id_list,
        )
        temp = {
            "seq_pro_score": float(0),
            "evalue": float(0),
            "domain_bias": float(0),
            "domain_score": float(0),
            "seq_pro_bias": float(0),
            "hmm_from": int(0),
            "hmm_to": int(0),
            "ali_from": int(0),
            "ali_to": int(0),
            "env_from": int(0),
            "env_to": int(0),
        }
        if get_protein_domains_result:
            for protein in get_protein_domains_result:
                for domain in protein["domains"]:
                    new_dict = temp | domain["domain_properties"]
                    new_dict["hmm_id"] = domain["hmm_id"]
                    self.proteins[protein["p1.uid"]].add_domain(
                        **new_dict,
                    )
        else:
            log.info(
                "None of the input protein ids were found or had domains in the database"
            )

    def annotate_proteins_with_neo4j(
        self,
        protein_uids: List[str] = None,
        chunk_size: int = 1000,
        annotate_all: bool = False,
        progress: bool = False,
    ):
        """
        The function `annotate_proteins_with_neo4j` takes a list of protein hash IDs, queries a database
        for matching proteins, and retrieves their HMM annotations.

        Args:
          protein_uids (List[str]): A list of protein hash IDs. These are unique identifiers for proteins in the database that you want to annotate.
          chunk_size (int): The `chunk_size` parameter determines the number of proteins to query at a time. Proteins are divided into chunks to improve efficiency when querying the database.  Defaults to 1000
          annotate_all (bool): The `annotate_all` parameter is a boolean flag that determines whether to annotate all proteins in the database or not. If set to `True`, all proteins in the database will be annotated. If set to `False`, you need to provide a list of protein hash IDs in the `protein_uids. Defaults to False
          progress (bool): The `progress` parameter is a boolean flag that determines whether or not to display a progress bar during the execution of the function. If `progress` is set to `True`, a progress bar will be displayed to track the progress of the function. If `progress` is set to `False`. Defaults to False

        """
        if protein_uids is None and annotate_all:
            protein_uids = self.get_all_feature_uids()
        elif isinstance(protein_uids, str):
            protein_uids = [protein_uids]
        log.info(
            f"Searching database for HMM annotations of {len(protein_uids)} proteins."
        )
        search_result = search_protein_hash(protein_uids)
        if not any([i for i in search_result.values()]):
            log.info("No identical proteins found in the database.")
        else:
            # log.info(
            #     f"{len([i for i in search_result.values() if i])} of {len(protein_uids)} searched proteins were found in the database, pulling their HMM annotations into python..."
            # )
            # get the number of chunks needed to to query "chunk_size" proteins at a time
            prot_len = len(protein_uids)
            n_chunks = prot_len // chunk_size
            if n_chunks == 0:
                n_chunks = 1
            if n_chunks > (len(protein_uids) - 1):
                chunked_list = chunk_a_list_with_numpy(
                    input_list=protein_uids, n_chunks=n_chunks
                )
            else:
                chunked_list = [protein_uids]
            del protein_uids
            if progress:
                with Progress(transient=True) as pg:
                    task = pg.add_task("Progress...", total=n_chunks)
                    for protein_id_list in chunked_list:
                        self.get_protein_domains_from_db(
                            protein_id_list=protein_id_list
                        )
                        pg.update(task, advance=1)
            else:
                for protein_id_list in chunked_list:
                    self.get_protein_domains_from_db(protein_id_list=protein_id_list)
        return search_result

    def annotate(
        self, use_neo4j_precalc: bool = False, neo4j_chunk_size: int = 1000, **kwargs
    ):
        """
        The `annotate` function is a convenience function that retrieves HMMER annotation for all
        proteins in a Socialgene object, either from a Neo4j database or by using HMMER directly.

        Args:
          use_neo4j_precalc (bool): A boolean flag indicating whether to use precalculated domain
        information from Neo4j for annotation. If set to True, the domains info will be retrieved from
        Neo4j, and proteins not found in Neo4j will be analyzed with HMMER. If set to False, all
        proteins will be analyzed with HMMER. Defaults to False
          neo4j_chunk_size (int): The `neo4j_chunk_size` parameter determines the number of proteins
        that will be sent to Neo4j at a time for annotation. Defaults to 1000
        """
        if use_neo4j_precalc:
            db_res = self.annotate_proteins_with_neo4j(
                annotate_all=True,
                chunk_size=neo4j_chunk_size,
            )
            n_not_in_db = [k for k, v in db_res.items() if not v]
        else:
            n_not_in_db = list(self.proteins.keys())
        if n_not_in_db:
            self.annotate_proteins_with_hmmscan(
                protein_id_list=n_not_in_db,
                **kwargs,
            )

    def annotate_proteins_with_hmmscan(
        self, protein_id_list, hmm_filepath, cpus=None, **kwargs
    ):
        """Run hmmscan on Protein objects

        Args:
            protein_id_list (list): list of protein hash ids to run hmmscan on
            hmm_filepath (str): path to file of HMMs
            cpus (int): number of parallel processes to spawn

        """
        log.info(f"Annotating {len(protein_id_list)} proteins with HMMER's hmmscan")
        if cpus is None:
            if cpu_count() > 2:
                cpus = cpu_count() - 1
            else:
                cpus = 1
        # create a list of proteins as FASTA
        temp1 = [
            self.proteins[i].fasta_string_defline_uid
            for i in protein_id_list
            if self.proteins[i].sequence
        ]
        if not temp1:
            log.info("None of the input sequences contain an amino acid sequence")
            return
        with tempfile.NamedTemporaryFile() as temp_path:
            hmmer_temp = HMMER(hmm_filepath=hmm_filepath)
            hmmer_temp.hmmscan(
                fasta_path="-",
                input="\n".join(temp1).encode(),
                domtblout_path=temp_path.name,
                overwrite=True,
                cpus=cpus,
                **kwargs,
            )
            # parse the resulting domtblout file, saving results to the class proteins/domains
            self.parse_hmmout(temp_path.name, hmmsearch_or_hmmscan="hmmscan")

    def get_proteins_from_neo4j(self):
        try:
            ncount = 0
            with GraphDriver() as db:
                for i in db.run(
                    """
                    MATCH (p1:protein)
                    RETURN p1.uid as uid
                    """,
                ):
                    _ = self.add_protein(uid=i["uid"])
                    ncount += 1
            log.info(f"Retrieved {ncount} proteins from Neo4j")
        except Exception:
            log.debug("Error trying to retrieve proteins from Neo4j")

    def add_sequences_from_neo4j(self):
        try:
            with GraphDriver() as db:
                for i in db.run(
                    """
                    MATCH (p1:protein)
                    WHERE p1.uid in $uid
                    RETURN p1.uid as uid, p1.sequence as sequence
                    """,
                    uid=[i.uid for i in self.protein_iter],
                ):
                    if i["uid"] in self.proteins:
                        self.proteins[i["uid"]].sequence = i["sequence"]
        except Exception:
            log.debug(f"Error trying to retrieve domains for {self.uid}")

    def hydrate_from_proteins(self):
        """Given a SocialGene object with proteins, retrieve from a running Neo4j database all locus and asssembly info for those proteins"""
        for result in Neo4jQuery.query_neo4j(
            cypher_name="retrieve_protein_locations",
            param=list(self.proteins.keys()),
        ):
            self.add_assembly(uid=result["assembly"], parent=self)
            for locus in result["loci"]:
                _ = self.assemblies[result["assembly"]].add_locus(
                    external_id=locus["locus"]
                )
                for feature in locus["features"]:
                    _ = (
                        self.assemblies[result["assembly"]]
                        .loci[locus["locus"]]
                        .add_feature(
                            type="protein",
                            uid=feature["external_id"],
                            start=feature["locus_start"],
                            end=feature["locus_end"],
                            strand=feature["strand"],
                        )
                    )

    def fill_given_locus_range(self, locus_uid, start, end):
        """Given a locus uid that's in the database, pull asssembly, locus, protein info for those proteins"""

        with GraphDriver() as db:
            assembly_uid = db.run(
                """
                MATCH (a1:assembly)<-[:ASSEMBLES_TO]-(n1:nucleotide {uid: $locus_uid})
                RETURN a1.uid as a_uid
            """,
                locus_uid=locus_uid,
            ).single()
        if not assembly_uid:
            raise ValueError("No assembly found in database")
        else:
            assembly_uid = assembly_uid.value()
        self.add_assembly(uid=assembly_uid, parent=self)
        with GraphDriver() as db:
            res = db.run(
                """
                MATCH (n1:nucleotide {uid: $locus_uid})
                RETURN n1 as nucleotide_node
            """,
                locus_uid=locus_uid,
            ).single()
        if not res:
            raise ValueError(f"{locus_uid} not found in database")
        external_id = res.value().get("external_id")
        self.assemblies[assembly_uid].add_locus(external_id=external_id)
        self.assemblies[assembly_uid].loci[external_id].metadata.update(
            dict(res.value())
        )

        with GraphDriver() as db:
            res = db.run(
                """
                MATCH (n1:nucleotide {uid: $locus_uid})-[e1:ENCODES]->(p1:protein)
                WHERE e1.start >= $start AND e1.start <= $end
                RETURN e1, p1
            """,
                locus_uid=locus_uid,
                start=start,
                end=end,
            )
            for feature in res:
                _ = self.add_protein(uid=feature.value().end_node["uid"])
                self.assemblies[assembly_uid].loci[external_id].add_feature(
                    type="protein",
                    uid=feature.value().end_node["uid"],
                    **feature.value(),
                )

        return {"assembly": assembly_uid, "locus": external_id}

    def _drop_all_cross_origin(self):
        # Drop likely cross-origin proteins (proteins that are longer than 100k bp)
        # have to remove from both gene cluster and locus
        for ak, av in self.assemblies.items():
            for nk, nv in av.loci.items():
                self.assemblies[ak].loci[nk].features = {
                    i for i in nv.features if abs(i.end - i.start) < 100000
                }
                for gene_cluster in self.assemblies[ak].loci[nk].gene_clusters:
                    gene_cluster.features = {
                        i
                        for i in gene_cluster.features
                        if abs(i.end - i.start) < 100000
                    }

    def hydrate_protein_info(self):
        """Pull name (original identifier) and description of proteins from Neo4j"""
        for result in Neo4jQuery.query_neo4j(
            cypher_name="get_protein_info",
            param=list(self.proteins.keys()),
        ):
            self.proteins[result["external_id"]].external_id = result["name"]
            self.proteins[result["external_id"]].description = result["description"]

    ########################################
    # File Outputs
    ########################################

    def export_all_domains_as_tsv(self, outpath, **kwargs):
        """
        The function exports all domains as a TSV file, sorted by protein ID and mean envelope position.

        Args:
          outpath: The `outpath` parameter is the path to the output file where the TSV (Tab-Separated
        Values) data will be written. It specifies the location and name of the file.
        """
        _domain_counter = 0
        with open_write(outpath, **kwargs) as f:
            tsv_writer = csv.writer(f, delimiter="\t")
            # sort to standardize the write order
            ordered_prot_ids = list(self.proteins.keys())
            ordered_prot_ids.sort()
            for prot_id in ordered_prot_ids:
                # sort to standardize the write order
                for domain in self.proteins[
                    prot_id
                ].domain_list_sorted_by_mean_envelope_position:
                    _domain_counter += 1
                    _temp = [prot_id]
                    _temp.extend(list(domain.all_attributes().values()))
                    tsv_writer.writerow(_temp)
        log.info(f"Wrote {str(_domain_counter)} domains to {outpath}")

    def ferment_pickle(self, outpath):
        """
        The function `ferment_pickle` saves a SocialGene object to a Python pickle file.

        Args:
          outpath: The `outpath` parameter is a string that represents the path where the pickled object will be saved. It should include the file name and extension. For example, "/path/to/save/object.pickle".
        """
        with open(outpath, "wb") as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def eat_pickle(inpath):
        """
        The `eat_pickle` function reads a saved SocialGene pickle file from the given path and returns a SocialGene object.

        Args:
          inpath: The `inpath` parameter is a string that represents the path to the file from which we
        want to load the pickled object.

        Returns:
          SocialGene object
        """
        with open(inpath, "rb") as handle:
            return pickle.load(handle)

    def filter_proteins(self, hash_list: List):
        """Filter proteins by list of hash ids

        Args:
            hash_list (List): List of protein hash identifiers

        Returns:
            Generator: generator returning tuple of length two -> Generator[(str, 'Protein'])
        """
        return ((k, v) for k, v in self.proteins.items() if k in hash_list)

    @property
    def fasta_string_defline_uid(self):
        for v in self.proteins.values():
            yield f">{v.uid}\n{v.sequence}\n"

    @property
    def fasta_string_defline_external_id(self):
        for v in self.proteins.values():
            yield f">{v.external_id}\n{v.sequence}\n"

    def write_fasta(
        self,
        outpath,
        external_id: bool = False,
        **kwargs,
    ):
        """Write all proteins to a FASTA file

        Args:
            outpath (str): path of file that FASTA entries will be appended to
            external_id (bool, optional): Write protein identifiers as the hash (True) or the original identifier (False). Defaults to False.
            **kwargs: see print(open_write.__doc__)
        """

        with open_write(filepath=outpath, **kwargs) as handle:
            counter = 0
            if external_id:
                fasta_gen = self.fasta_string_defline_external_id
            else:
                fasta_gen = self.fasta_string_defline_uid
            for i in fasta_gen:
                counter += 1
                if "compression" in kwargs and kwargs["compression"]:
                    handle.write(i.encode())
                else:
                    handle.writelines(i)

        log.info(f"Wrote {str(counter)} proteins to {outpath}")

    def write_n_fasta(self, outdir, n_splits=1, **kwargs):
        """
        The function `write_n_fasta` exports protein sequences split into multiple fasta files.

        Args:
          outdir: The `outdir` parameter is a string that specifies the directory where the fasta files
        will be saved.
          n_splits: The `n_splits` parameter in the `write_n_fasta` function determines the number of
        fasta files the protein sequences will be split into. By default, it is set to 1, meaning all
        the protein sequences will be written into a single fasta file. If you specify a value greater
        than. Defaults to 1
        """

        # this can be done with itertools.batched in python 3.12
        def split(a, n):
            # https://stackoverflow.com/a/2135920
            k, m = divmod(len(a), n)
            return (
                a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)
            )

        protein_list = split(
            [value for key, value in sorted(self.proteins.items(), reverse=False)],
            n_splits,
        )
        counter = 1
        for protein_group in protein_list:
            with open_write(
                Path(outdir, f"fasta_split_{counter}.faa"), **kwargs
            ) as handle:
                for i in protein_group:
                    handle.writelines(f">{i.uid}\n{i.sequence}\n")
            counter += 1

    def _merge_proteins(self, sg_object):
        """
        The function merges the proteins from another SOcialGene object into the current object.

        Args:
          sg_object: The `sg_object` parameter is an object of the same class as the current object. It
        represents another object that contains a collection of proteins.
        """
        self.proteins.update(sg_object.proteins)

    def _merge_assemblies(self, sg_object):
        """
        The function `_merge_assemblies` merges assemblies and loci from `sg_object` into `self` if they
        do not already exist, and adds protein features to loci if they already exist.

        Args:
          sg_object: The `sg_object` parameter is an object of the same class as the current object.
        """
        for assembly_k, assembly_v in sg_object.assemblies.items():
            if assembly_k not in self.assemblies:
                self.assemblies[assembly_k] = assembly_v
            else:
                for locus_k, locus_v in assembly_v.loci.items():
                    if locus_k not in self.assemblies[assembly_k].loci:
                        self.assemblies[assembly_k].loci[locus_k] = locus_v
                    else:
                        for feature in (
                            sg_object.assemblies[assembly_k].loci[locus_k].features
                        ):
                            self.assemblies[assembly_k].loci[locus_k].add_feature(
                                type="protein",
                                uid=feature.uid,
                                start=feature.start,
                                end=feature.end,
                                strand=feature.strand,
                            )

    # def deep_merge_with_sg1(sg_object_1, sg_object_2):
    def __add__(self, sg_object):
        """
        The function merges proteins and assemblies from two objects and appends protein comparison data
        to a list or dataframe.

        Args:
          sg_object: The `sg_object` parameter is an object of the same class as the current object. It
        represents another instance of the class that you want to add to the current instance.
        """
        self._merge_proteins(sg_object)
        self._merge_assemblies(sg_object)
        self.protein_comparison.extend(sg_object.protein_comparison)

    def write_genbank(self, outpath, compression=None):
        for assembly in self.assemblies.values():
            for locus in assembly.loci.values():
                record = SeqRecord(
                    Seq(""),
                    id=locus.external_id,
                    name=locus.external_id,
                    description="A GenBank file generated by SocialGene.",
                    dbxrefs=[f"Assembly:{locus.parent.uid}"],
                )
                # Add annotation
                for feature in locus.features_sorted_by_midpoint:
                    biofeat = SeqFeature(
                        FeatureLocation(
                            start=feature.start - 1,
                            end=feature.end,
                            strand=feature.strand,
                        ),
                        type=feature.type,
                        qualifiers={
                            k: v
                            for k, v in feature.all_attributes().items()
                            if v and k not in ["parent", "protein"]
                        }
                        | {"translation": self.proteins[feature.uid].sequence},
                    )
                    record.features.append(biofeat)
                record.annotations["molecule_type"] = "DNA"
                with open_write(outpath, "a", compression) as h:
                    SeqIO.write(
                        record,
                        h,
                        "genbank",
                    )

    def drop_all_non_protein_features(self, **kwargs):
        """Drop features from all assembly/loci that aren't proteins/pseudo-proteins

        Returns:
            list: list of features that were removed (if return_removed=True)
        """
        for a_v in self.assemblies.values():
            for l_v in a_v.loci.values():
                l_v.drop_non_protein_features()

    ########################################
    # OUTPUTS FOR NEXTFLOW
    ########################################

    def write_table(
        self,
        outdir: str,
        tablename: str,
        filename: str = None,
        include_sequences=False,
        **kwargs,
    ):
        """
        The function writes a table to a specified output directory in TSV format, for import into Neo4j.

        Args:
          outdir (str): The `outdir` parameter is a string that specifies the directory where the output
        file will be saved.
          tablename (str): The `tablename` parameter is a string that specifies the name of the table to
        be written.
          filename (str): The `filename` parameter is an optional argument that specifies the name of
        the file to be written. If no `filename` is provided, the `tablename` will be used as the
        filename.
        """
        if not filename:
            filename = tablename
        outpath = Path(outdir, filename)
        with open_write(outpath, **kwargs) as handle:
            tsv_writer = csv.writer(
                handle, delimiter="\t", quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            for i in getattr(self, tablename)(include_sequences=include_sequences):
                tsv_writer.writerow(i)

    def table_protein_to_go(self, **kwargs):
        """
        The function `table_protein_to_go` iterates through the assemblies, loci, and features of a
        given object, and yields tuples containing the protein hash and GO term for each feature that
        has GO terms, for import into Neo4j.
        """
        for av in self.assemblies.values():
            for v in av.loci.values():
                for feature in v.features:
                    # not all features will have goterms so check here
                    if feature.goterms:
                        for goterm in feature.goterms:
                            yield (
                                feature.uid,
                                goterm.removeprefix("GO:").strip(),
                            )

    def table_locus_to_protein(self, **kwargs):
        """
        The function `table_locus_to_protein` generates a table of protein information from each locus, for import into Neo4j.
        """
        for ak, av in self.assemblies.items():
            for k, loci in av.loci.items():
                temp_list = list(loci.features)
                # sort features by id then start to maintain consistent output
                temp_list.sort(key=attrgetter("uid"))
                temp_list.sort(key=attrgetter("start"))
                for feature in temp_list:
                    if feature.feature_is_protein():
                        yield (
                            feature.parent.uid,
                            feature.uid,
                            feature.external_id,
                            feature.locus_tag,
                            feature.start,
                            feature.end,
                            feature.strand,
                            feature.description,
                            feature.partial_on_complete_genome,
                            feature.missing_start,
                            feature.missing_stop,
                            feature.internal_stop,
                            feature.partial_in_the_middle_of_a_contig,
                            feature.missing_N_terminus,
                            feature.missing_C_terminus,
                            feature.frameshifted,
                            feature.too_short_partial_abutting_assembly_gap,
                            feature.incomplete,
                        )

    def table_protein_ids(self, include_sequences=False, **kwargs):
        """Protein hash id table for import into Neo4j"""
        for protein in self.proteins.values():
            if include_sequences:
                yield (protein.uid, protein.crc64, protein.sequence)
            else:
                yield (protein.uid, protein.crc64)

    def table_assembly_to_locus(self, **kwargs):
        """Assembly to locus table for import into Neo4j

        Args:
            outdir (str, optional): Defaults to ".".
        """
        for ak, av in self.assemblies.items():
            for k, v in av.loci.items():
                #  ["assembly", "internal_locus_id"]
                yield (ak, v.uid)

    def table_loci(self, **kwargs):
        """
        Generate a table of loci information.

        Yields:
            tuple: (internal_locus_id, external_locus_id, [info])
        """
        for _av in self.assemblies.values():
            for locus in _av.loci.values():
                yield tuple(
                    [locus.uid]
                    + [locus.external_id]
                    + list(locus.metadata.all_attributes().values())
                )

    def table_assembly(self, **kwargs):
        """Assembly table for import into Neo4j

        Args:
            outdir (str, optional): Defaults to ".".
        """
        for assembly in self.assemblies.values():
            yield tuple(
                [assembly.uid] + list(assembly.metadata.all_attributes().values())
            )

    def table_assembly_to_taxid(self, **kwargs):
        """Assembly table for import into Neo4j

        Args:
            outdir (str, optional): Defaults to ".".
        """
        for k, v in self.assemblies.items():
            if v.taxid:
                yield (k, v.taxid)



import subprocess

from socialgene.utils.logging import CONSOLE, log


def run_subprocess(
    command_list,
    check=True,
    shell=False,
    capture_output=True,
    status=False,
    **kwargs,
):
    """Run something in a separate process

    Args:
        command_list (list or str): list if shell=False, str if shell=True # TODO:better preventative
        check (bool, optional): [description]. Defaults to True.
        shell (bool, optional): [description]. Defaults to False.

    Raises:
        ValueError: [description]
    """
    if not command_list:
        log.error("Attempted to pass an empty command to subprocess.run()")
        raise ValueError
    if isinstance(command_list, list):
        command_list_string = " ".join(command_list)
    else:
        command_list_string = command_list
    if status:
        log.info(f"Executing external program:\n{command_list_string}")
        with CONSOLE.status(
            "",
            spinner="bouncingBar",
        ) as status:
            result = subprocess.run(
                command_list,
                check=check,
                shell=shell,
                capture_output=capture_output,
                **kwargs,
            )
    else:
        log.debug(f"Executing external program:\n{command_list_string}")
        result = subprocess.run(
            command_list,
            check=check,
            shell=shell,
            capture_output=capture_output,
            **kwargs,
        )
    if result.returncode != 0:
        log.error(f"Error code: {result.returncode}")
        log.error(f"Error stdout: {result.stderr.decode('utf-8')}")
        raise SystemExit



import datetime

import numpy as np


def np_json_converter(obj):
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, datetime.datetime):
        return obj.__str__()



from types import GeneratorType

from socialgene.compare_proteins.diamond import DiamondBlastp
from socialgene.compare_proteins.hmmer import CompareDomains
from socialgene.compare_proteins.mmseqs import MMseqsEasySearch


class BGCComparison:
    def __init__(self, tool):
        match tool:
            case "blastp":
                self.tool = DiamondBlastp()
            case "mmseqs2":
                self.tool = MMseqsEasySearch()
            case "hmmer":
                self.tool = CompareDomains()
            case _:
                raise ValueError(
                    f"Tool {tool} not recognised, must be one of 'blastp', 'mmseqs2', 'hmmer'"
                )

    def compare(self, bgc1, bgc2, **kwargs):
        if isinstance(bgc1, GeneratorType):
            bgc1 = list(bgc1)
        if isinstance(bgc2, GeneratorType):
            bgc2 = list(bgc2)
        return self.tool.reciprocal_hits(
            self.tool.compare_proteins(bgc1, bgc2, **kwargs),
            self.tool.compare_proteins(bgc2, bgc1, **kwargs),
        )



import sys
from importlib.metadata import entry_points


def find_entrypoints():
    entrypoints = [
        i.name
        for i in entry_points(group="console_scripts")
        if i.value.startswith("socialgene")
    ]
    entrypoints.sort()
    return entrypoints


def main():
    print("Socialgene has the following entry points for direct command line use:")
    print("\n".join(find_entrypoints()))


if __name__ == "__main__":
    main()
    sys.exit(main() or 0)



import argparse
import multiprocessing
from functools import partial
from itertools import batched

import pandas as pd
from rdkit import DataStructs

from socialgene.addons.chemistry.nr import ChemicalCompoundNode, TanimotoSimilarity
from socialgene.base.chem import ChemicalCompound
from socialgene.neo4j.neo4j import GraphDriver

cmpd_label = ChemicalCompoundNode.neo4j_label[0]


def create_arg_parser():
    parser = argparse.ArgumentParser(
        description="Calculate chemical similarity between all the chemical nodes in a SocialGene Neo4j Database"
    )
    parser.add_argument(
        "--cpus",
        help="Number of cpus to use, memory usage will increase linearly. Default is 1.",
        required=False,
        default=1,
        type=int,
    )
    parser.add_argument(
        "--threshold",
        help="Tanimoto similarity threshold, values equal to or above this will be included in the database. Default is 0.5.",
        required=False,
        default=0.5,
        type=float,
    )
    return parser


def get_db_inchis():
    with GraphDriver() as db:
        res = db.run(
            f"""
            MATCH (c1:{cmpd_label})
            RETURN c1.CanonicalSmiles as chem
            """,
        ).value()
    return [i.replace("\\\\", "\\") for i in res if i is not None]


def inchi_list_to_compound_dict(x):
    res = {}
    for i in x:
        temp = ChemicalCompound(i, sanitize=True)
        res[i] = (temp, temp.node)
    return res


def create_inchi_list_to_compound_dict(inchlist, cpus=1):
    group_size = len(inchlist) // cpus
    b = batched(inchlist, group_size)
    combined_dict = {}
    with multiprocessing.Pool(processes=cpus) as pool:
        for res in pool.imap_unordered(inchi_list_to_compound_dict, b):
            combined_dict.update(res)
    return combined_dict


def calculate_similarity(tupgroup, threshold=0.5):
    chemsim_set = set()
    tup = tupgroup[2]
    for iter in range(tupgroup[0], tupgroup[1]):
        for i, x in enumerate(DataStructs.BulkTanimotoSimilarity(tup[iter], tup)):
            if x >= threshold:
                if tup[iter] != i:
                    score = int(x * 100)
                    temp = (iter, i, score)
                    chemsim_set.add(temp)
    return chemsim_set


def calculate_similarity_parallel(combined_dict, threshold=0.5, cpus=1):
    id_list = list(combined_dict.keys())
    big_vec = [i[0].morgan for i in combined_dict.values()]
    group_size = len(id_list) // cpus
    groups = [(i, i + group_size, big_vec) for i in range(0, len(id_list), group_size)]
    groups[-1] = (groups[-1][0], len(id_list) - 1, big_vec)
    chemsim = set()
    with multiprocessing.Pool(processes=cpus) as pool:
        calculate_similarity_partial = partial(
            calculate_similarity, threshold=threshold
        )
        for i in pool.imap_unordered(calculate_similarity_partial, groups):
            chemsim.update(i)
    chemsim2 = set()
    for i in chemsim:
        # sorted prevents identical forward + reverse relationships
        chemsim2.add(tuple(sorted(i[0:2]) + [i[2]]))
    return chemsim2


def combine(combined_dict, chemsim):
    df0 = pd.DataFrame(
        ((k, v[0], v[1]) for k, v in combined_dict.items()),
        columns=["id", "compound", "node"],
    )
    df = pd.DataFrame(chemsim, columns=["start", "end", "similarity"])
    df = df.set_index("start")
    df = df.join(df0["node"])
    df.columns = ["end", "similarity", "start_node"]
    df = df.reset_index(level=0)
    df = df.set_index("end")
    df = df.join(df0["node"])
    df.columns = ["start", "similarity", "start_node", "end_node"]
    df = df.reset_index(level=0)
    df["relationship"] = df.apply(
        lambda x: TanimotoSimilarity(
            start=x.start_node, end=x.end_node, properties={"similarity": x.similarity}
        ),
        axis=1,
    )
    return df


def main():
    parser = create_arg_parser()
    args = parser.parse_args()
    if args.cpus == 1:
        print("Warning: Running with 1 cpu will be slow")
    inchis = get_db_inchis()
    combined_dict = create_inchi_list_to_compound_dict(inchis, cpus=args.cpus)
    chemsim = calculate_similarity_parallel(
        combined_dict, threshold=args.threshold, cpus=args.cpus
    )
    df = combine(combined_dict, chemsim)
    del combined_dict, chemsim
    # remove self relationships
    df = df[df.start != df.end]
    z = df["relationship"].tolist()
    z[0].add_multiple_to_neo4j(z, create=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import pickle
import sys
from pathlib import Path

from socialgene.cli.search.sea import search_bgc


def run_individual_search(gbk_path):
    outpath_clinker = Path(
        "/media/socialgene_nvme/culture_search_results/jsons", gbk_path.stem + ".json"
    )
    if (
        not outpath_clinker.exists()
        and not Path(
            "/media/socialgene_nvme/culture_search_results/errors",
            gbk_path.stem + ".error",
        ).exists()
    ):
        try:
            a = search_bgc(
                input=gbk_path,
                hmm_dir="/media/socialgene_nvme/v0.4.1/refseq/socialgene_per_run/hmm_cache",
                outpath_clinker=outpath_clinker,
                use_neo4j_precalc=True,
                assemblies_must_have_x_matches=0.4,
                nucleotide_sequences_must_have_x_matches=0.4,
                gene_clusters_must_have_x_matches=0.4,
                break_bgc_on_gap_of=10000,
                target_bgc_padding=10000,
                max_domains_per_protein=3,
                max_outdegree=300000,
                max_query_proteins=50,
                scatter=True,
                locus_tag_bypass_list=None,
                protein_id_bypass_list=None,
                only_culture_collection=True,
                frac=0.75,
                run_async=True,
                analyze_with="blastp",
                blast_speed="ultra-sensitive",
            )
            with open(
                Path(
                    "/media/socialgene_nvme/culture_search_results/pickles",
                    gbk_path.stem + ".pickle",
                ),
                "wb",
            ) as handle:
                pickle.dump(a, handle, protocol=pickle.HIGHEST_PROTOCOL)
        except Exception as e:
            with open(
                Path(
                    "/media/socialgene_nvme/culture_search_results/errors",
                    gbk_path.stem + ".error",
                ),
                "w",
            ) as handle:
                handle.write(str(e))


if __name__ == "__main__":
    # run run_individual_search() using a command line input as the gbk_path
    gbk_path = Path(sys.argv[1])
    outdir = "/media/socialgene_nvme/culture_search_results"

    Path(outdir).mkdir(exist_ok=True)
    Path(outdir, "jsons").mkdir(exist_ok=True)
    Path(outdir, "pickles").mkdir(exist_ok=True)
    Path(outdir, "errors").mkdir(exist_ok=True)
    run_individual_search(gbk_path)



import re
from abc import ABC
from itertools import batched
from typing import List

from socialgene.neo4j.neo4j import GraphDriver
from socialgene.utils.logging import log


class Neo4jElement(ABC):
    neo4j_label = None  # list for nodes because they can have multiple labels, string for relationships because they can only have one
    description = None
    property_specification = {}
    header = None
    header_filename = None
    target_subdirectory = None
    target_extension = None
    required_properties = None

    def __init__(
        self,
        properties: dict = None,
        **kwargs,
    ):
        """This defines nodes and relationships that will be imported  into Neo4j  with the Neo4j Admin Import tool
        For more information, especially about what makes up the headers, see: https://neo4j.com/docs/operations-manual/current/tutorial/neo4j-admin-import/

        Args:
            neo4j_label (str): this will become the Neo4j node or relationship LABEL
            description (str): helpful information about what the element represents
            header_filename (str): the name of the header file used for Neo4j admin import (basically column names, but not quite)
            target_subdirectory (str): subdirectory the import data can be found in  (e.g. for non-redundant protein nodes it would be 'protein_info' because data is within: `$outdir/socialgene_neo4j/import/protein_info`)
            target_extension (str): extension that is unique to the wanted data files (scoped within target_subdirectory)
            header (List): list of strings, each string will make up a column in a tab separated header file for Neo4j admin import (basically column names, but not quite)
        """
        self.override = False
        # properties aren't required for relationships
        if self.property_specification is None:
            self.property_specification = {}
        # if not provided then all properties are required

        if self.required_properties is None:
            self.required_properties = list(self.property_specification.keys())
        if self.required_properties is None:
            self.required_properties = []
        if properties is None:
            self.properties = {}
        if properties is not None:
            self.properties = properties
            self._clean_properties()

    @property
    def __required_properties_dict(self):
        return {i: self.santized_properties.get(i) for i in self.required_properties}

    @property
    def __optional_properties_dict(self):
        return {
            i: self.santized_properties.get(i)
            for i in self.properties
            if i not in self.required_properties
        }

    def __hash__(self):
        return hash((self.neo4j_label, frozenset(self.santized_properties.items())))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.neo4j_label == other.neo4j_label:
                if self.santized_properties == other.santized_properties:
                    return True
        return False

    def _check_required_properties(
        self,
    ):
        missing = [i for i in self.required_properties if i not in self.properties]
        if missing:
            raise ValueError(
                f"{self.__class__} was missing required properties: {missing}"
            )

    def _check_no_extra_properties(self):
        extra = [i for i in self.properties if i not in self.property_specification]
        if extra:
            raise ValueError(
                f"{self.__class__} contained unsupported extra properties: {extra}"
            )

    def _check_property_types(self):
        bad = {
            k: f"expected {self.property_specification.get(k).__name__} but got {type(v).__name__}"
            for k, v in self.properties.items()
            if not isinstance(v, self.property_specification.get(k))
        }
        if bad:
            raise ValueError(
                f"{self.__class__} contained properties with incorrect types: {bad}"
            )
        self.properties = {
            k: v
            for k, v in self.properties.items()
            if isinstance(v, self.property_specification.get(k))
        }

    @property
    def santized_properties(self):
        """Remove any properties that are None"""
        temp = {}
        for k, v in self.properties.items():
            if isinstance(v, str):
                v = v.replace("\\", "\\\\")
            temp[k] = v
        return temp

    def _clean_properties(
        self,
    ):
        # only keep properties that are in the property specification and remove nulls
        self.properties = {
            k: v
            for k, v in self.properties.items()
            if k in self.property_specification and v is not None
        }
        if self.override:
            return
        self._check_required_properties()
        self._check_no_extra_properties()
        self._check_property_types()

    def fill_from_dict(self, d: dict):
        """Fill the properties of the object from a dictionary"""
        for k, v in self.property_specification.items():
            try:
                if k not in d:
                    log.debug(
                        f"Failed to fill add property {k} from dictionary: not found for {self.__class__}"
                    )
                    continue
                if d.get(k) is None:
                    log.debug(
                        f"Failed to fill add property {k} from dictionary: None for {self.__class__}"
                    )
                    continue

                self.properties[k] = v(d.get(k))
            except Exception as e:
                log.error(
                    f"Failed to fill add property {k} from dictionary: {e} for {self.__class__}"
                )


class Node(Neo4jElement):
    """Represents a single Node"""

    nonunique_index = None

    def __init__(
        self,
        uid: List[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if self.property_specification is None:
            raise ValueError(
                f"property_specification must be defined for class {self.__class__.__name__}"
            )
        if isinstance(self.neo4j_label, str):
            self.neo4j_label = [self.neo4j_label]

    def __hash__(self):
        return hash(
            (frozenset(self.neo4j_label), frozenset(self.santized_properties.items()))
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.neo4j_label == other.neo4j_label:
                if self.santized_properties == other.santized_properties:
                    return True
        return False

    def __str__(self) -> str:
        return f'(:{":".join(self.neo4j_label)})'

    def _neo4j_repr(
        self,
        var="n",
    ):
        """Create a string of the form (var:LABEL {uid: 'uid', ...})"""
        uids = {i: self.santized_properties.get(i) for i in self._Node__uid}
        uids_as_str = ", ".join(
            f"{k}: '{v}'" if isinstance(v, str) else f"{k}: {v}"
            for k, v in uids.items()
        )
        return f"({var}: {self.neo4j_label} {{{uids_as_str}}})"

    def _neo4j_repr_params(
        self,
        var="n",
        map_key="required_props",
    ):
        """Return a string of the form (var:LABEL {uid: $uid, ...})
        Ensure that the required properties are included in the string
        """
        req_props = self._Neo4jElement__required_properties_dict.keys()
        req_props = [f"{i}: {map_key}.{i}" for i in req_props]
        req_props = ", ".join(req_props)
        return f"({var}: {self.neo4j_label[0]} {{{req_props}}})"

    def add_constraints_to_neo4j(self):
        """Add constraints to Neo4j"""
        if not hasattr(self, "constraints_unique") or not self.constraints_unique:
            log.info(f"No unique constraints to add for {self.__str__()} nodes")
            return
        # for now at least only use the first label
        label = self.neo4j_label[0]
        with GraphDriver() as db:
            res = db.run(
                f"""
                CREATE CONSTRAINT python_added_index_{label} IF NOT EXISTS
                FOR (n:{label})
                REQUIRE ({", ".join([f"n.{i}" for i in self.constraints_unique])}) IS UNIQUE
                """
            ).consume()
            try:
                # The code is checking if the variable `res` has a property called `counters` and if
                # it does, it will execute the code block inside the `if` statement.
                if res.counters:
                    log.info(
                        f"Created {res.counters.constraints_added} constraints for {self.__str__()} nodes"
                    )
            except Exception:
                pass

    def add_nonunique_index_to_neo4j(self):
        """Add non-unique indices to Neo4j"""
        if not hasattr(self, "nonunique_index") or not self.nonunique_index:
            log.info(f"No non-unique indices to add for {self.__str__()} nodes")
            return
        label = self.neo4j_label[0]
        with GraphDriver() as db:
            res = db.run(
                f"""
                CREATE INDEX python_added_non_unique_index_{label} IF NOT EXISTS
                FOR (n:{label})
                ON ({", ".join([f"n.{i}" for i in self.nonunique_index])})
                """
            ).consume()
            try:
                if res.counters:
                    log.info(
                        f"Created {res.counters.indexes_added} non-unique indices for {self.__str__()} nodes"
                    )
            except Exception:
                pass

    def add_to_neo4j(self, create=False):
        """Add a single node to Neo4j"""
        var = "n"
        merge_str = self._neo4j_repr_params(var=var, map_key="required_props")
        paramsdict = {
            "optional_props": self._Neo4jElement__optional_properties_dict,
            "required_props": self._Neo4jElement__required_properties_dict,
        }
        if len(self.neo4j_label) > 1:
            # handle adding extra labels if they exist
            add_label_str = ":".join(self.neo4j_label[1:])
            add_label_str = f"SET {var}:{add_label_str}"
        else:
            add_label_str = ""
        if create:
            with GraphDriver() as db:
                res = db.run(
                    f"""
                    WITH $paramsdict as paramsdict
                    WITH paramsdict.optional_props as optional_props, paramsdict.required_props as required_props
                    CREATE {merge_str}
                    SET n += optional_props
                    {add_label_str}
                    """,
                    paramsdict=paramsdict,
                ).consume()
                try:
                    log.info(
                        f"Created {res.counters.nodes_created} {self.__str__()} nodes, set {res.counters.properties_set} properties"
                    )
                except Exception:
                    pass
        else:
            if len(self.neo4j_label) > 1:
                add_label_str = ":".join(self.neo4j_label[1:])
                add_label_str = f"ON CREATE SET {var}:{add_label_str}"
            else:
                add_label_str = ""
            with GraphDriver() as db:
                res = db.run(
                    f"""
                    WITH $paramsdict as paramsdict
                    WITH paramsdict.optional_props as optional_props, paramsdict.required_props as required_props
                    MERGE {merge_str}
                    ON CREATE SET n += optional_props
                    {add_label_str}
                    """,
                    paramsdict=paramsdict,
                ).consume()
                try:
                    log.info(
                        f"Created {res.counters.nodes_created} {self.__str__()} nodes, set {res.counters.properties_set} properties"
                    )
                except Exception:
                    pass

    def update_properties(self):
        """Update the properties of a node"""
        var = "n"
        merge_str = self._neo4j_repr_params(var=var, map_key="required_props")
        paramsdict = {
            "optional_props": self._Neo4jElement__optional_properties_dict,
            "required_props": self._Neo4jElement__required_properties_dict,
        }
        if len(self.neo4j_label) > 1:
            # handle adding extra labels if they exist
            add_label_str = ":".join(self.neo4j_label[1:])
            add_label_str = f"SET {var}:{add_label_str}"
        else:
            add_label_str = ""
        with GraphDriver() as db:
            res = db.run(
                f"""
                    WITH $paramsdict as paramsdict
                    WITH paramsdict.optional_props as optional_props, paramsdict.required_props as required_props
                    MATCH {merge_str}
                    SET n += optional_props
                    {add_label_str}
                    """,
                paramsdict=paramsdict,
            ).consume()
            try:
                log.info(
                    f"Created {res.counters.nodes_created} {self.__str__()} nodes, set {res.counters.properties_set} properties"
                )
            except Exception:
                pass

    @staticmethod
    def update_properties_many(list_of_nodes, batch_size=1000):
        """Update the properties of a node"""
        """Add multiple nodes to Neo4j at a time in transactions of batch_size nodes at a time"""
        if not all(
            isinstance(sub, type(list_of_nodes[0])) for sub in list_of_nodes[1:]
        ):
            raise ValueError(
                f"All nodes in list_of_nodes must be of the same type as the first element, found (not in order): {set([sub.__class__.__name__ for sub in list_of_nodes])}"
            )
        single = list_of_nodes[0]
        count_res = {"nodes_created": 0, "properties_set": 0}
        var = "n"
        merge_str = single._neo4j_repr_params(var=var, map_key="required_props")
        if len(single.neo4j_label) > 1:
            # handle adding extra labels if they exist
            add_label_str = ":".join(single.neo4j_label[1:])
        else:
            add_label_str = ""
        for paramsdictlist in batched(
            (
                {
                    "optional_props": i._Neo4jElement__optional_properties_dict,
                    "required_props": i._Neo4jElement__required_properties_dict,
                }
                for i in list_of_nodes
            ),
            batch_size,
        ):
            with GraphDriver() as db:
                res = db.run(
                    f"""
                    WITH $paramsdictlist as paramsdictlist
                    UNWIND paramsdictlist as paramsdict
                    WITH paramsdict.optional_props as optional_props, paramsdict.required_props as required_props
                    MATCH {merge_str}
                    SET n += optional_props
                    {add_label_str}
                    """,
                    paramsdictlist=paramsdictlist,
                ).consume()
                try:
                    count_res["nodes_created"] += res.counters.nodes_created
                    count_res["properties_set"] += res.counters.properties_set
                except Exception:
                    pass
        log.info(
            f"Created {count_res['nodes_created']} {single.__str__()} nodes, set {count_res['properties_set']} properties"
        )

    @staticmethod
    def add_multiple_to_neo4j(list_of_nodes, batch_size=1000, create=False):
        """Add multiple nodes to Neo4j at a time in transactions of batch_size nodes at a time"""
        if not all(
            isinstance(sub, type(list_of_nodes[0])) for sub in list_of_nodes[1:]
        ):
            raise ValueError(
                f"All nodes in list_of_nodes must be of the same type as the first element, found (not in order): {set([sub.__class__.__name__ for sub in list_of_nodes])}"
            )
        single = list_of_nodes[0]
        count_res = {"nodes_created": 0, "properties_set": 0}
        var = "n"
        merge_str = single._neo4j_repr_params(var=var, map_key="required_props")
        if len(single.neo4j_label) > 1:
            # handle adding extra labels if they exist
            add_label_str = ":".join(single.neo4j_label[1:])
            if create:
                add_label_str = f"SET {var}:{add_label_str}"
            else:
                add_label_str = f"ON CREATE SET {var}:{add_label_str}"
        else:
            add_label_str = ""
        for paramsdictlist in batched(
            (
                {
                    "optional_props": i._Neo4jElement__optional_properties_dict,
                    "required_props": i._Neo4jElement__required_properties_dict,
                }
                for i in list_of_nodes
            ),
            batch_size,
        ):
            if create:
                with GraphDriver() as db:
                    res = db.run(
                        f"""
                        WITH $paramsdictlist as paramsdictlist
                        UNWIND paramsdictlist as paramsdict
                        WITH paramsdict.optional_props as optional_props, paramsdict.required_props as required_props
                        CREATE {merge_str}
                        SET n += optional_props
                        {add_label_str}
                        """,
                        paramsdictlist=paramsdictlist,
                    ).consume()
                try:
                    count_res["nodes_created"] += res.counters.nodes_created
                    count_res["properties_set"] += res.counters.properties_set
                except Exception:
                    pass
            else:
                with GraphDriver() as db:
                    res = db.run(
                        f"""
                        WITH $paramsdictlist as paramsdictlist
                        UNWIND paramsdictlist as paramsdict
                        WITH paramsdict.optional_props as optional_props, paramsdict.required_props as required_props
                        MERGE {merge_str}
                        ON CREATE SET n += optional_props
                        {add_label_str}
                        """,
                        paramsdictlist=paramsdictlist,
                    ).consume()
                try:
                    count_res["nodes_created"] += res.counters.nodes_created
                    count_res["properties_set"] += res.counters.properties_set
                except Exception:
                    pass
        log.info(
            f"Created {count_res['nodes_created']} {single.__str__()} nodes, set {count_res['properties_set']} properties"
        )


class Relationship(Neo4jElement):
    def __init__(self, start=None, end=None, **kwargs):
        super().__init__(**kwargs)
        self.start = start
        self.end = end
        if not isinstance(self.start, self.start_class):
            raise ValueError(
                f"Relationship class '{self.__class__.__name__}' start node must be of type {self.start_class.__name__}, found {self.start.__class__.__name__}"
            )
        if not isinstance(self.end, self.end_class):
            raise ValueError(
                f"Relationship class '{self.__class__.__name__}' end node must be of type {self.end_class.__name__}, found {self.end.__class__.__name__}"
            )

    def __str__(self):
        return f"{self.start.__str__()}-[:{self.neo4j_label}]->{self.end.__str__()}"

    def __hash__(self):
        return hash(
            (
                self.neo4j_label,
                frozenset(self.santized_properties.items()),
                self.start,
                self.end,
            )
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.neo4j_label == other.neo4j_label:
                if self.santized_properties == other.santized_properties:
                    if self.start == other.start:
                        if self.end == other.end:
                            return True
        return False

    @property
    def neo4j_label_str(self):
        return str(self.neo4j_label)

    def _start_field(self):
        for i in self.header:
            if "START_ID" in i:
                return i

    def _end_field(self):
        for i in self.header:
            if "END_ID" in i:
                return i

    @property
    def _extract_start_from_admin_header(self):
        try:
            return re.findall(r"\((.*?)\)", self._start_field())[0]
        except Exception as e:
            raise e

    @property
    def _extract_end_from_admin_header(self):
        try:
            return re.findall(r"\((.*?)\)", self._end_field())[0]
        except Exception as e:
            raise e

    @property
    def _cypher_string(self):
        return f"(:{self.start_class.neo4j_label})-[:{self.neo4j_label}]->(:{self.end_class.neo4j_label})"

    def add_to_neo4j(self, create=False):
        match_start = self.start._neo4j_repr_params(var="m1", map_key="required_props1")
        match_end = self.end._neo4j_repr_params(var="m2", map_key="required_props2")
        # self.start.add_to_neo4j()
        # self.end.add_to_neo4j()
        the_json = {
            "required_props1": self.start._Neo4jElement__required_properties_dict,
            "required_props2": self.end._Neo4jElement__required_properties_dict,
            "properties": self.santized_properties,
        }
        if create:
            with GraphDriver() as db:
                x = db.run(
                    f"""
                    WITH $properties as properties
                    WITH properties.required_props1 as required_props1, properties.required_props2 as required_props2, properties.properties as props
                    MATCH {match_start}
                    MATCH {match_end}
                    CALL apoc.create.relationship(m1, '{self.neo4j_label}', props, m2)
                    YIELD rel
                    RETURN count(rel) as relationships_created
                    """,
                    properties=the_json,
                ).value()
        else:
            with GraphDriver() as db:
                x = db.run(
                    f"""
                    WITH $properties as properties
                    WITH properties.required_props1 as required_props1, properties.required_props2 as required_props2, properties.properties as props
                    MATCH {match_start}
                    MATCH {match_end}
                    CALL apoc.merge.relationship(m1, '{self.neo4j_label}', props, {{}}, m2, {{}})
                    YIELD rel
                    RETURN count(rel) as relationships_created
                    """,
                    properties=the_json,
                ).value()
        log.info(f"{x[0]} relationships created {self.__str__()}")

    @staticmethod
    def add_multiple_to_neo4j(list_of_rels, batch_size=1000, create=False):
        """Add multiple relationships to Neo4j at a time in transactions of batch_size nodes at a time"""
        if not all(isinstance(sub, type(list_of_rels[0])) for sub in list_of_rels[1:]):
            raise ValueError(
                f"All relationships in list_of_rels must be of the same type as the first element, found (not in order): {set([sub.__class__.__name__ for sub in list_of_rels])}"
            )
        single = list_of_rels[0]
        match_start = single.start._neo4j_repr_params(
            var="m1", map_key="required_props1"
        )
        match_end = single.end._neo4j_repr_params(var="m2", map_key="required_props2")
        n_rels_created = 0
        for paramsdictlist in batched(
            (
                {
                    "required_props1": i.start._Neo4jElement__required_properties_dict,
                    "required_props2": i.end._Neo4jElement__required_properties_dict,
                    "properties": i.santized_properties,
                }
                for i in list_of_rels
            ),
            batch_size,
        ):
            if create:
                with GraphDriver() as db:
                    x = db.run(
                        f"""
                        WITH $paramsdictlist as paramsdictlist
                        UNWIND paramsdictlist as properties
                        WITH properties.required_props1 as required_props1, properties.required_props2 as required_props2, properties.properties as props
                        MATCH {match_start}
                        MATCH {match_end}
                        CALL apoc.create.relationship(m1, '{single.neo4j_label}', props, m2 )
                        YIELD rel
                        RETURN count(rel) as relationships_created
                        """,
                        paramsdictlist=paramsdictlist,
                    ).value()
                n_rels_created += x[0]
            else:
                with GraphDriver() as db:
                    x = db.run(
                        f"""
                        WITH $paramsdictlist as paramsdictlist
                        UNWIND paramsdictlist as properties
                        WITH properties.required_props1 as required_props1, properties.required_props2 as required_props2, properties.properties as props
                        MATCH {match_start}
                        MATCH {match_end}
                        CALL apoc.merge.relationship(m1, '{single.neo4j_label}', props, {{}}, m2, {{}})
                        YIELD rel
                        RETURN count(rel) as relationships_created
                        """,
                        paramsdictlist=paramsdictlist,
                    ).value()
                n_rels_created += x[0]
        log.info(f"{n_rels_created} relationships created {single.__str__()}")



import pandas as pd

from socialgene.base.socialgene import SocialGene
from socialgene.cli.search.sea import search_bgc
from socialgene.neo4j.neo4j import GraphDriver
from socialgene.utils.logging import log

if __name__ == "__main__":

    with GraphDriver() as db:
        bgcs = db.run(
            """
            MATCH (n:nucleotide)
            WHERE n.external_id STARTS WITH "BGC"
            RETURN n.uid as uid
            """,
        ).value()

    counter = 0
    threshold = 0.7

    for bgc in bgcs:
        counter += 1
        log.warning(f"Running {counter} of {len(bgcs)}")
        sg = SocialGene()
        sg.fill_given_locus_range(locus_uid=bgc, start=float("-inf"), end=float("inf"))
        a = search_bgc(
            input=sg,
            hmm_dir="/home/chase/Documents/socialgene_data/v0_4_1/refseq_antismash_bgcs/socialgene_per_run/hmm_cache",
            outpath_clinker=None,
            use_neo4j_precalc=True,
            assemblies_must_have_x_matches=threshold,
            nucleotide_sequences_must_have_x_matches=threshold,
            gene_clusters_must_have_x_matches=threshold,
            break_bgc_on_gap_of=20000,
            target_bgc_padding=20000,
            max_domains_per_protein=3,
            max_outdegree=500000,
            max_query_proteins=10,
            scatter=False,
            locus_tag_bypass_list=None,
            protein_id_bypass_list=None,
            only_culture_collection=False,
            frac=0.75,
            run_async=True,
            analyze_with="blastp",
        )
        temp = pd.merge(
            a._compare_bgcs_by_jaccard_and_levenshtein(),
            a._compare_bgcs_by_median_bitscore(),
            left_on="query_gene_cluster",
            right_on="target_gene_cluster",
            how="inner",
        ).sort_values(by=["modscore", "score"], ascending=False)

        temp["nuc"] = temp["query_gene_cluster_x"].apply(lambda x: x.parent.uid)
        # get best hit per nuc sequence
        temp = temp.groupby("nuc").head(1).reset_index(drop=True)
        temp = temp[
            ["query_gene_cluster_x", "jaccard", "levenshtein", "modscore", "score"]
        ]
        temp = temp[temp["jaccard"] > threshold]
        temp["start"] = temp["query_gene_cluster_x"].apply(
            lambda x: min(i.start for i in x.features)
        )
        temp["end"] = temp["query_gene_cluster_x"].apply(
            lambda x: max(i.end for i in x.features)
        )
        temp["query_gene_cluster_x"] = temp["query_gene_cluster_x"].apply(
            lambda x: x.parent.uid
        )
        # remove rows if query_gene_cluster_x is the same as bgc
        temp = temp[temp["query_gene_cluster_x"] != bgc]
        temp = temp.to_dict(orient="records")
        if temp:
            log.info(f"Linking {len(temp)} assemblies to {bgc}")
            with GraphDriver() as db:
                db.run(
                    """
                    MATCH (n:nucleotide)
                    WHERE n.uid=$bgc_id
                    UNWIND $to_link as to_link
                    MATCH (m:nucleotide)
                    WHERE m.uid=to_link['query_gene_cluster_x']
                    MERGE (n)-[r:LINKS_TO]->(m)
                        ON CREATE SET r.jaccard=to_link['jaccard'], r.levenshtein=to_link['levenshtein'], r.modscore=to_link['modscore'], r.score=to_link['score'], r.start=to_link['start'], r.end=to_link['end']
                    """,
                    bgc_id=bgc,
                    to_link=temp,
                )



#
# from pathlib import Path

#

#
# from socialgene.utils.printing import (
#     print_test_name,
#     print_message,
#     print_success,
#     print_error,
# )


# def check_hashed_file(neo4j_outdir, subpath, expected_files):
#     print_test_name("Check tigrfam files:")
#     inspect_dir = Path(neo4j_outdir, *subpath)
#     print_message("Looking for files in:")
#     print_message(str(inspect_dir))
#     filenames = [i.name for i in Path.iterdir(inspect_dir)]
#     # separate for-loops so the different cases print out next to each other
#     for i in expected_files:
#         if i in filenames:
#             print_success(f"exists:  {i}")  # extra space aligns print with missing
#     for i in expected_files:
#         if i not in filenames:
#             print_error(f"missing: {i}")
#     for i in filenames:
#         if i not in expected_files:
#             print_error(f"extra:   {i}")


# def check_for_header(neo4j_outdir, expected_headers):
#     print_test_name("Check for expected header files:")
#     import_directory = Path(neo4j_outdir, "neo4j", "import", "headers")
#     print_message("Looking for directories in:")
#     print_message(str(import_directory))
#     for i in Path.iterdir(import_directory):
#         if i.name in expected_headers:
#             print_success(f"exists:  {i.name}")  # extra space aligns print with missing
#         else:
#             print_error(f"missing: {i.name}")



"""https://www.ebi.ac.uk/chebi/"""

from socialgene.neo4j.neo4j_element import Node


class ChebiNode(Node):
    neo4j_label = ["chebi"]
    description = "Represents a ChEBI term"
    property_specification = {
        "uid": int,
        "name": str,
    }
    constraints_unique = ["uid"]



#
# from collections import OrderedDict
# from pathlib import Path
# import json
# import csv

#

#
# from socialgene.utils.logging import log


# class DatasetsParse:
#     pass


# class Biosample:

#     biosample_attributes = [
#         "biosample_accession",
#         "collection_date",
#         "culture_collection",
#         "env_broad_scale",
#         "env_local_scale",
#         "env_medium",
#         "env_package",
#         "geo_loc_name",
#         "host",
#         "investigation_type",
#         "isol_growth_condt",
#         "isolation_source",
#         "lat_lon",
#         "num_replicons",
#         "project_name",
#         "ref_biomaterial",
#         "sample_type",
#         "type-material",
#     ]

#     def __init__(self, jsonl_path, outdir):
#         self.jsonl_path = jsonl_path
#         self.outdir = outdir
#         self.jsonl_dict = None

#         self._validate()
#         # self._load_jsonl()
#         self._parse_biosamples()

#     def _validate(self):
#         for i in ["assembly_to_biosample", "biosamples"]:
#             temp_path = Path(self.outdir, i)
#             if temp_path.exists() and temp_path.stat().st_size > 0:
#                 raise IOError(
#                     f"File exists already and isn't empty, taking the coward's way out and aborting: {temp_path}"
#                 )

#     def _get_all_biosample_attribute_names(self):
#         # this may be unnecessary but I don't know if the fields change between assemblies
#         self.biosample_attributes = set()
#         for i in self.jsonl_dict:
#             for att in i["assemblyInfo"]["biosample"]["attributes"]:
#                 self.biosample_attributes.add(att["name"])
#         self.biosample_attributes = list(self.biosample_attributes)
#         self.biosample_attributes.sort()
#         # the results of self.biosample_attributes is hardcoded as a dictionary for ensured consistency

#     def _parse_biosamples(self):
#         # File reading/writing here top prevent having data copies everywhere

#         with open(self.jsonl_path) as input_data:
#             for line in input_data:
#                 jsonl_dict = json.loads(line)
#                 with open(
#                     Path(self.outdir, "assembly_to_biosample"), "a"
#                 ) as assembly_to_biosample_handle, open(
#                     Path(self.outdir, "biosamples"), "a"
#                 ) as biosamples_handle:
#                     assembly_to_biosample_tsv_writer = csv.writer(
#                         assembly_to_biosample_handle,
#                         delimiter="\t",
#                         quotechar='"',
#                         quoting=csv.QUOTE_MINIMAL,
#                     )
#                     biosamples_tsv_writer = csv.writer(
#                         biosamples_handle,
#                         delimiter="\t",
#                         quotechar='"',
#                         quoting=csv.QUOTE_MINIMAL,
#                     )

#                     assembly_to_biosample_tsv_writer.writerow(
#                         [
#                             jsonl_dict["assemblyInfo"]["assemblyAccession"],
#                             jsonl_dict["assemblyInfo"]["biosample"]["accession"],
#                         ]
#                     )
#                     biosample_attribute_dict = OrderedDict(
#                         {i: None for i in self.biosample_attributes}
#                     )
#                     biosample_attribute_dict["biosample_accession"] = jsonl_dict[
#                         "assemblyInfo"
#                     ]["biosample"]["accession"]
#                     for att in jsonl_dict["assemblyInfo"]["biosample"]["attributes"]:
#                         try:
#                             temp = att["value"]
#                             temp.replace("\t", " ")
#                             if not att["value"] == "missing":
#                                 biosample_attribute_dict[att["name"]] = att["value"]
#                             if att["name"] == "culture_collection":
#                                 try:
#                                     temp = att["value"].split
#                                 except Exception as e:
#                                     log.error(e)
#                         except Exception as e:
#                             log.debug(f"{att['name']} not found in dict")
#                             log.error(e)
#                             pass
#                     log.info(list(biosample_attribute_dict.values()))
#                     biosamples_tsv_writer.writerow(
#                         list(biosample_attribute_dict.values())
#                     )



import argparse
import sys

from socialgene.addons.gnps_networking.parse import GNPS_SNETS
from socialgene.utils.logging import log


def create_arg_parser():
    parser = argparse.ArgumentParser(
        description="Integrate a NPAtlas molecular network into a SocialGene Neo4j Database"
    )
    parser.add_argument(
        "--gnps_dirpath",
        help="Local path to unzipped GNPS Molecular Networking download",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--params_xml_path",
        help="Local path to params.xml file",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--specnets_path",
        help="Local path to specnets file",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--selfloop_path",
        help="Local path to selfloop file",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--clustersummary_path",
        help="Local path to clustersummary file",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--clusterinfo_path",
        help="Local path to clusterinfo file",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--map_path",
        help="Local path to map file (tab separated file with two columns and no header; the first column is the assembly id and the second column is the GNPS mass spec filename)",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--force",
        help="Force integration of GNPS molecular network into the database",
        action="store_true",
        default=False,
        required=False,
    )
    return parser


def main():
    parser = create_arg_parser()
    args = parser.parse_args()
    if args.gnps_dirpath:
        gnps = GNPS_SNETS(gnps_dirpath=args.gnps_dirpath, map_path=args.map_path)
        gnps._find_gnps_paths()
    else:
        # check args must contain all the paths to the GNPS files and they must exist
        for i in [
            args.params_xml_path,
            args.specnets_path,
            args.selfloop_path,
            args.clustersummary_path,
            args.clusterinfo_path,
            # args.map_path,
        ]:
            if not i:
                log.error(
                    "If --gnps_dirpath is not provided, then all GNPS file paths must be provided"
                )
                return
        gnps = GNPS_SNETS(
            params_xml_path=args.params_xml_path,
            specnets_path=args.specnets_path,
            selfloop_path=args.selfloop_path,
            clustersummary_path=args.clustersummary_path,
            clusterinfo_path=args.clusterinfo_path,
            map_path=args.map_path,
        )
    gnps._parse()
    if not args.force:
        if args.map_path:
            # assembly_check = gnps._validate_map()
            # if assembly_check.shape[0] > 0:
            #     if not args.force:
            #         raise ValueError(
            #             "Not all assemblies in the GNPS network were found in the database or vice-versa.\nTo continue anyways, use the --force flag."
            #         )
            ...
        else:
            log.error(
                "To link MS data to genomes, provide --map_path; to skip this step, use the --force flag."
            )
            sys.exit(1)

    # loop through all the methods
    for i in [
        "_add_input_spectra",
        "_create_library_hit_nodes",
        "_add_library_hit_nodes",
        "_add_library_hit_nodes",
        "_add_cluster_nodes",
        "_add_input_ms_files",
        "_link_cluster_to_spectrum",
        "_link_spectrum_to_ms_file",
        "_link_npclassifiers",
        "_link_cluster_to_library_no_spec_id",
        "_link_cluster_to_library",
        "_link_network",
        "_link_library_to_chem",
    ]:
        try:
            getattr(gnps, i)()
        except Exception as e:
            log.error(f"Failed to {i}: {e}")

    if args.map_path:
        try:
            gnps._link_ms_file_to_assembly()
        except Exception as e:
            log.error(f"Failed to link mass spec file to assembly: {e}")

    # gnps.add_input_spectra_to_gnps_clusters()
    # gnps.add_links_between_gnps_clusters()
    log.info(
        "GNPS molecular network has been integrated into the SocialGene Neo4j database"
    )


if __name__ == "__main__":
    main()



import argparse
import concurrent.futures
import json
import tempfile
from collections import defaultdict
from pathlib import Path

from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn

from socialgene.addons.npatlas.parse import NPAtlasEntry, _download_npatlas
from socialgene.utils.logging import log


def create_arg_parser():
    """ "Creates and returns the ArgumentParser object."""
    parser = argparse.ArgumentParser(
        description="Integrate NPAtlas into a SocialGene Neo4j Database"
    )
    parser.add_argument(
        "--input",
        help="Local path to NPAtlas JSON file. If None, downloads from NPAtlas website.",
        default=None,
        required=False,
    )
    return parser


def main():
    parser = create_arg_parser()
    args = parser.parse_args()
    input = args.input
    if input is None:
        with tempfile.NamedTemporaryFile() as temp:
            Path(temp.name).exists()
            _download_npatlas(outpath=temp.name)
            with open(temp.name) as f:
                entries = json.load(f)
    else:
        jpath = Path(input)
        if jpath.exists():
            with open(jpath) as f:
                entries = json.load(f)
        else:
            raise FileNotFoundError(f"Path {input} does not exist")

    nodes = set()
    rels = defaultdict(set)

    def process_entry(entry):
        z = NPAtlasEntry(entry)
        z.parse()
        return z

    with Progress(
        TextColumn("{task.completed}"),
        "[progress.description]{task.description}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        TimeElapsedColumn(),
    ) as pg:
        task = pg.add_task("[cyan]Processing NPAtlas entries...", total=33372)
        # This was multithreaded whcih worked well on the first iteraion of this code when it was primarily reading
        # but after chem data calculation was added, it may be just as slow as single
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for entry in concurrent.futures.as_completed(
                [executor.submit(process_entry, entry) for entry in entries]
            ):
                z = entry.result()
                nodes.add(z.node)
                for k, v in z.get_links().items():
                    rels[k].update(v)
                pg.update(task, advance=1)
    log.info("Creating/Merging npatlas nodes in neo4")
    list(nodes)[0].add_multiple_to_neo4j(list(nodes), create=False)
    relnodes = defaultdict(set)
    for k, v in rels.items():
        for i in v:
            if i.start.__class__.__name__ != "NPAtlasNode":
                relnodes[i.start.__class__].add(i.start)
            if i.end.__class__.__name__ != "NPAtlasNode":
                relnodes[i.end.__class__].add(i.end)

    log.info("Creating/Merging nodes linked to npatlas entries in neo4j")
    # make sure nodes that npatlas entries are linked to exist in the database
    for k, v in relnodes.items():
        list(v)[0].add_multiple_to_neo4j(list(v), create=False)
    log.info("Linking npatlas entries and related nodes in neo4j")
    for k, v in rels.items():
        k.add_multiple_to_neo4j(list(v), create=False)


if __name__ == "__main__":
    main()



import argparse
from pathlib import Path

from socialgene.hmm.hmminfo import HmmInfo

parser = argparse.ArgumentParser(description="Merging tsv")
parser.add_argument(
    "--all_hmms_path",
    help=".sg_hmm_nodes filepath",
    required=True,
)


def main():
    args = parser.parse_args()
    all_hmms_path = Path(args.all_hmms_path)
    a = HmmInfo(all_hmms_path)
    a.read_tsv()
    a.write_nr_hmm_nodes(outpath="sg_hmm_nodes")
    a.write_hmm_source_nodes(outdir=".")


if __name__ == "__main__":
    main()
